"""Resource provider tests."""

import decimal
import sys
from typing import Any

from dependency_injector import containers, providers, resources, errors
from pytest import raises, mark


def init_fn(*args, **kwargs):
    return args, kwargs


def test_is_provider():
    assert providers.is_provider(providers.Resource(init_fn)) is True


def test_init_optional_provides():
    provider = providers.Resource()
    provider.set_provides(init_fn)
    assert provider.provides is init_fn
    assert provider() == (tuple(), dict())


def test_set_provides_returns_():
    provider = providers.Resource()
    assert provider.set_provides(init_fn) is provider


@mark.parametrize(
    "str_name,cls",
    [
        ("dependency_injector.providers.Factory", providers.Factory),
        ("decimal.Decimal", decimal.Decimal),
        ("list", list),
        (".test_resource_py35.test_is_provider", test_is_provider),
        ("test_is_provider", test_is_provider),
    ],
)
def test_set_provides_string_imports(str_name, cls):
    assert providers.Resource(str_name).provides is cls


def test_provided_instance_provider():
    provider = providers.Resource(init_fn)
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_injection():
    resource = object()

    def _init():
        _init.counter += 1
        return resource

    _init.counter = 0

    class Container(containers.DeclarativeContainer):
        resource = providers.Resource(_init)
        dependency1 = providers.List(resource)
        dependency2 = providers.List(resource)

    container = Container()
    list1 = container.dependency1()
    list2 = container.dependency2()

    assert list1 == [resource]
    assert list1[0] is resource

    assert list2 == [resource]
    assert list2[0] is resource

    assert _init.counter == 1


def test_init_function():
    def _init():
        _init.counter += 1

    _init.counter = 0

    provider = providers.Resource(_init)

    result1 = provider()
    assert result1 is None
    assert _init.counter == 1

    result2 = provider()
    assert result2 is None
    assert _init.counter == 1

    provider.shutdown()


def test_init_generator():
    def _init():
        _init.init_counter += 1
        yield
        _init.shutdown_counter += 1

    _init.init_counter = 0
    _init.shutdown_counter = 0

    provider = providers.Resource(_init)

    result1 = provider()
    assert result1 is None
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 0

    provider.shutdown()
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 1

    result2 = provider()
    assert result2 is None
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 1

    provider.shutdown()
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 2


def test_init_class():
    class TestResource(resources.Resource):
        init_counter = 0
        shutdown_counter = 0

        def init(self):
            self.__class__.init_counter += 1

        def shutdown(self, _):
            self.__class__.shutdown_counter += 1

    provider = providers.Resource(TestResource)

    result1 = provider()
    assert result1 is None
    assert TestResource.init_counter == 1
    assert TestResource.shutdown_counter == 0

    provider.shutdown()
    assert TestResource.init_counter == 1
    assert TestResource.shutdown_counter == 1

    result2 = provider()
    assert result2 is None
    assert TestResource.init_counter == 2
    assert TestResource.shutdown_counter == 1

    provider.shutdown()
    assert TestResource.init_counter == 2
    assert TestResource.shutdown_counter == 2


def test_init_class_generic_typing():
    # See issue: https://github.com/ets-labs/python-dependency-injector/issues/488
    class TestDependency:
        ...

    class TestResource(resources.Resource[TestDependency]):
        def init(self, *args: Any, **kwargs: Any) -> TestDependency:
            return TestDependency()

        def shutdown(self, resource: TestDependency) -> None: ...

    assert issubclass(TestResource, resources.Resource) is True


def test_init_class_abc_init_definition_is_required():
    class TestResource(resources.Resource):
        ...

    with raises(TypeError) as context:
        TestResource()

    assert "Can't instantiate abstract class TestResource" in str(context.value)
    assert "init" in str(context.value)


def test_init_class_abc_shutdown_definition_is_not_required():
    class TestResource(resources.Resource):
        def init(self):
            ...

    assert hasattr(TestResource(), "shutdown") is True


def test_init_not_callable():
    provider = providers.Resource(1)
    with raises(errors.Error):
        provider.init()


def test_init_and_shutdown():
    def _init():
        _init.init_counter += 1
        yield
        _init.shutdown_counter += 1

    _init.init_counter = 0
    _init.shutdown_counter = 0

    provider = providers.Resource(_init)

    result1 = provider.init()
    assert result1 is None
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 0

    provider.shutdown()
    assert _init.init_counter == 1
    assert _init.shutdown_counter == 1

    result2 = provider.init()
    assert result2 is None
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 1

    provider.shutdown()
    assert _init.init_counter == 2
    assert _init.shutdown_counter == 2


def test_shutdown_of_not_initialized():
    def _init():
        yield

    provider = providers.Resource(_init)

    result = provider.shutdown()
    assert result is None


def test_initialized():
    provider = providers.Resource(init_fn)
    assert provider.initialized is False

    provider.init()
    assert provider.initialized is True

    provider.shutdown()
    assert provider.initialized is False


def test_call_with_context_args():
    provider = providers.Resource(init_fn, "i1", "i2")
    assert provider("i3", i4=4) == (("i1", "i2", "i3"), {"i4": 4})


def test_fluent_interface():
    provider = providers.Resource(init_fn) \
        .add_args(1, 2) \
        .add_kwargs(a3=3, a4=4)
    assert provider() == ((1, 2), {"a3": 3, "a4": 4})


def test_set_args():
    provider = providers.Resource(init_fn) \
        .add_args(1, 2) \
        .set_args(3, 4)
    assert provider.args == (3, 4)


def test_clear_args():
    provider = providers.Resource(init_fn) \
        .add_args(1, 2) \
        .clear_args()
    assert provider.args == tuple()


def test_set_kwargs():
    provider = providers.Resource(init_fn) \
        .add_kwargs(a1="i1", a2="i2") \
        .set_kwargs(a3="i3", a4="i4")
    assert provider.kwargs == {"a3": "i3", "a4": "i4"}


def test_clear_kwargs():
    provider = providers.Resource(init_fn) \
        .add_kwargs(a1="i1", a2="i2") \
        .clear_kwargs()
    assert provider.kwargs == {}


def test_call_overridden():
    provider = providers.Resource(init_fn, 1)
    overriding_provider1 = providers.Resource(init_fn, 2)
    overriding_provider2 = providers.Resource(init_fn, 3)

    provider.override(overriding_provider1)
    provider.override(overriding_provider2)

    instance1 = provider()
    instance2 = provider()

    assert instance1 is instance2
    assert instance1 == ((3,), {})
    assert instance2 == ((3,), {})


def test_deepcopy():
    provider = providers.Resource(init_fn, 1, 2, a3=3, a4=4)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert provider.args == provider_copy.args
    assert provider.kwargs == provider_copy.kwargs
    assert isinstance(provider, providers.Resource)


def test_deepcopy_initialized():
    provider = providers.Resource(init_fn)
    provider.init()

    with raises(errors.Error):
        providers.deepcopy(provider)


def test_deepcopy_from_memo():
    provider = providers.Resource(init_fn)
    provider_copy_memo = providers.Resource(init_fn)

    provider_copy = providers.deepcopy(
        provider,
        memo={id(provider): provider_copy_memo},
    )

    assert provider_copy is provider_copy_memo


def test_deepcopy_args():
    provider = providers.Resource(init_fn)
    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider.add_args(dependent_provider1, dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.args[0]
    dependent_provider_copy2 = provider_copy.args[1]

    assert provider.args != provider_copy.args

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_kwargs():
    provider = providers.Resource(init_fn)
    dependent_provider1 = providers.Factory(list)
    dependent_provider2 = providers.Factory(dict)

    provider.add_kwargs(d1=dependent_provider1, d2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.kwargs["d1"]
    dependent_provider_copy2 = provider_copy.kwargs["d2"]

    assert provider.kwargs != provider_copy.kwargs

    assert dependent_provider1.cls is dependent_provider_copy1.cls
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.cls is dependent_provider_copy2.cls
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_overridden():
    provider = providers.Resource(init_fn)
    object_provider = providers.Object(object())

    provider.override(object_provider)

    provider_copy = providers.deepcopy(provider)
    object_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert provider.args == provider_copy.args
    assert isinstance(provider, providers.Resource)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


def test_deepcopy_with_sys_streams():
    provider = providers.Resource(init_fn)
    provider.add_args(sys.stdin, sys.stdout, sys.stderr)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider_copy, providers.Resource)
    assert provider.args[0] is sys.stdin
    assert provider.args[1] is sys.stdout
    assert provider.args[2] is sys.stderr


def test_repr():
    provider = providers.Resource(init_fn)

    assert repr(provider) == (
        "<dependency_injector.providers.Resource({0}) at {1}>".format(
            repr(init_fn),
            hex(id(provider)),
        )
    )



"""Tests for wiring with dynamic container."""

from dependency_injector import containers, providers
from pytest import fixture

from samples.wiringstringids import module, package
from samples.wiringstringids.service import Service


@fixture(autouse=True)
def container():
    sub = containers.DynamicContainer()
    sub.int_object = providers.Object(1)

    container = containers.DynamicContainer()
    container.config = providers.Configuration()
    container.service = providers.Factory(Service)
    container.sub = sub

    container.wire(
        modules=[module],
        packages=[package],
    )

    yield container

    container.unwire()


def test_wire():
    service = module.test_function()
    assert isinstance(service, Service)



"""Tests for compatibility of @inject-patched functions with asyncio and inspect module checks."""

import asyncio
import inspect

from dependency_injector.wiring import inject


def test_asyncio_iscoroutinefunction():
    @inject
    async def foo():
        ...

    assert asyncio.iscoroutinefunction(foo)


def test_inspect_iscoroutinefunction():
    @inject
    async def foo():
        ...

    assert inspect.iscoroutinefunction(foo)



"""ProvidedInstance provider async mode tests."""

import asyncio

from dependency_injector import containers, providers
from pytest import mark, raises

from .common import RESOURCE1, init_resource


@mark.asyncio
async def test_provided_attribute():
    class TestClient:
        def __init__(self, resource):
            self.resource = resource

    class TestService:
        def __init__(self, resource):
            self.resource = resource

    class TestContainer(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
        client = providers.Factory(TestClient, resource=resource)
        service = providers.Factory(TestService, resource=client.provided.resource)

    container = TestContainer()

    instance1, instance2 = await asyncio.gather(
        container.service(),
        container.service(),
    )

    assert instance1.resource is RESOURCE1
    assert instance2.resource is RESOURCE1
    assert instance1.resource is instance2.resource


@mark.asyncio
async def test_provided_attribute_error():
    async def raise_exception():
        raise RuntimeError()

    class TestContainer(containers.DeclarativeContainer):
        client = providers.Factory(raise_exception)

    container = TestContainer()

    with raises(RuntimeError):
        await container.client.provided.attr()


@mark.asyncio
async def test_provided_attribute_undefined_attribute():
    class TestClient:
        def __init__(self, resource):
            self.resource = resource

    class TestContainer(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
        client = providers.Factory(TestClient, resource=resource)

    container = TestContainer()

    with raises(AttributeError):
        await container.client.provided.attr()


@mark.asyncio
async def test_provided_item():
    class TestClient:
        def __init__(self, resource):
            self.resource = resource

        def __getitem__(self, item):
            return getattr(self, item)

    class TestService:
        def __init__(self, resource):
            self.resource = resource

    class TestContainer(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
        client = providers.Factory(TestClient, resource=resource)
        service = providers.Factory(TestService, resource=client.provided["resource"])

    container = TestContainer()

    instance1, instance2 = await asyncio.gather(
        container.service(),
        container.service(),
    )

    assert instance1.resource is RESOURCE1
    assert instance2.resource is RESOURCE1
    assert instance1.resource is instance2.resource


@mark.asyncio
async def test_provided_item_error():
    async def raise_exception():
        raise RuntimeError()

    class TestContainer(containers.DeclarativeContainer):
        client = providers.Factory(raise_exception)

    container = TestContainer()

    with raises(RuntimeError):
        await container.client.provided["item"]()


@mark.asyncio
async def test_provided_item_undefined_item():
    class TestContainer(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
        client = providers.Factory(dict, resource=resource)

    container = TestContainer()

    with raises(KeyError):
        await container.client.provided["item"]()


@mark.asyncio
async def test_provided_method_call():
    class TestClient:
        def __init__(self, resource):
            self.resource = resource

        def get_resource(self):
            return self.resource

    class TestService:
        def __init__(self, resource):
            self.resource = resource

    class TestContainer(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource, providers.Object(RESOURCE1))
        client = providers.Factory(TestClient, resource=resource)
        service = providers.Factory(TestService, resource=client.provided.get_resource.call())

    container = TestContainer()

    instance1, instance2 = await asyncio.gather(
        container.service(),
        container.service(),
    )

    assert instance1.resource is RESOURCE1
    assert instance2.resource is RESOURCE1
    assert instance1.resource is instance2.resource


@mark.asyncio
async def test_provided_method_call_parent_error():
    async def raise_exception():
        raise RuntimeError()

    class TestContainer(containers.DeclarativeContainer):
        client = providers.Factory(raise_exception)

    container = TestContainer()

    with raises(RuntimeError):
        await container.client.provided.method.call()()


@mark.asyncio
async def test_provided_method_call_error():
    class TestClient:
        def method(self):
            raise RuntimeError()

    class TestContainer(containers.DeclarativeContainer):
        client = providers.Factory(TestClient)

    container = TestContainer()

    with raises(RuntimeError):
        await container.client.provided.method.call()()



"""Object provider tests."""

from dependency_injector import providers


def test_is_provider():
    assert providers.is_provider(providers.Object(object())) is True


def test_init_optional_provides():
    instance = object()
    provider = providers.Object()
    provider.set_provides(instance)
    assert provider.provides is instance
    assert provider() is instance


def test_set_provides_returns_():
    provider = providers.Object()
    assert provider.set_provides(object()) is provider


def test_provided_instance_provider():
    provider = providers.Object(object())
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_call_object_provider():
    obj = object()
    assert providers.Object(obj)() is obj


def test_call_overridden_object_provider():
    obj1 = object()
    obj2 = object()
    provider = providers.Object(obj1)
    provider.override(providers.Object(obj2))
    assert provider() is obj2


def test_deepcopy():
    provider = providers.Object(1)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider, providers.Object)


def test_deepcopy_from_memo():
    provider = providers.Object(1)
    provider_copy_memo = providers.Provider()

    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})

    assert provider_copy is provider_copy_memo


def test_deepcopy_overridden():
    provider = providers.Object(1)
    overriding_provider = providers.Provider()

    provider.override(overriding_provider)

    provider_copy = providers.deepcopy(provider)
    overriding_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert isinstance(provider, providers.Object)

    assert overriding_provider is not overriding_provider_copy
    assert isinstance(overriding_provider_copy, providers.Provider)


def test_deepcopy_doesnt_copy_provided_object():
    # Fixes bug #231
    # Details: https://github.com/ets-labs/python-dependency-injector/issues/231
    some_object = object()
    provider = providers.Object(some_object)

    provider_copy = providers.deepcopy(provider)

    assert provider() is some_object
    assert provider_copy() is some_object


def test_repr():
    some_object = object()
    provider = providers.Object(some_object)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Object({0}) at {1}>".format(repr(some_object), hex(id(provider)))
    )



import pytest
from dependency_injector.providers import ThreadLocalSingleton
from dependency_injector.containers import Container

class FailingClass:
    def __init__(self):
        raise ValueError("FAILING CLASS")

class TestContainer(Container):
    failing_class = ThreadLocalSingleton(FailingClass)

def test_on_failure_value_error_is_raised():
    container = TestContainer()

    with pytest.raises(ValueError, match="FAILING CLASS"):
        container.failing_class()


"""Provider utils tests."""

from dependency_injector import providers


def test_with_instance():
    assert providers.is_provider(providers.Provider()) is True


def test_with_class():
    assert providers.is_provider(providers.Provider) is False


def test_with_string():
    assert providers.is_provider("some_string") is False


def test_with_object():
    assert providers.is_provider(object()) is False


def test_with_subclass_instance():
    class SomeProvider(providers.Provider):
        pass

    assert providers.is_provider(SomeProvider()) is True


def test_with_class_with_getattr():
    class SomeClass(object):
        def __getattr__(self, _):
            return False

    assert providers.is_provider(SomeClass()) is False



"""Named injection tests."""

from dependency_injector import providers


def test_isinstance():
    injection = providers.NamedInjection("name", 1)
    assert isinstance(injection, providers.Injection)


def test_get_name():
    injection = providers.NamedInjection("name", 123)
    assert injection.get_name() == "name"


def test_get_value_with_not_provider():
    injection = providers.NamedInjection("name", 123)
    assert injection.get_value() == 123


def test_get_value_with_factory():
    injection = providers.NamedInjection("name", providers.Factory(object))

    obj1 = injection.get_value()
    obj2 = injection.get_value()

    assert type(obj1) is object
    assert type(obj2) is object
    assert obj1 is not obj2


def test_get_original_value():
    provider = providers.Factory(object)
    injection = providers.NamedInjection("name", provider)
    assert injection.get_original_value() is provider


def test_deepcopy():
    provider = providers.Factory(object)
    injection = providers.NamedInjection("name", provider)

    injection_copy = providers.deepcopy(injection)

    assert injection_copy is not injection
    assert injection_copy.get_original_value() is not injection.get_original_value()

def test_deepcopy_memo():
    provider = providers.Factory(object)
    injection = providers.NamedInjection("name", provider)
    injection_copy_orig = providers.NamedInjection("name", provider)

    injection_copy = providers.deepcopy(injection, {id(injection): injection_copy_orig})

    assert injection_copy is injection_copy_orig
    assert injection_copy.get_original_value() is injection.get_original_value()



"""Tests for container config loading."""

from dependency_injector import containers, providers
from pytest import fixture


@fixture
def yaml_config_file(tmp_path):
    config_file = str(tmp_path / "config.yml")
    with open(config_file, "w") as file:
        file.write(
            "section1:\n"
            "  value1: yaml-loaded\n"
        )
    return config_file


@fixture
def ini_config_file(tmp_path):
    config_file = str(tmp_path / "config.ini")
    with open(config_file, "w") as file:
        file.write(
            "[section2]:\n"
            "value2 = ini-loaded\n"
        )
    return config_file



def test_auto_load(yaml_config_file, ini_config_file):
    class ContainerWithConfig(containers.DeclarativeContainer):
        config = providers.Configuration(
            yaml_files=[yaml_config_file],
            ini_files=[ini_config_file],
        )

    container = ContainerWithConfig()
    assert container.config.section1.value1() == "yaml-loaded"
    assert container.config.section2.value2() == "ini-loaded"


def test_auto_load_and_overriding(yaml_config_file):
    class ContainerWithConfig(containers.DeclarativeContainer):
        config = providers.Configuration(yaml_files=[yaml_config_file])

    container = ContainerWithConfig(config={"section1": {"value1": "overridden"}})
    assert container.config.section1.value1() == "overridden"


def test_manual_load(yaml_config_file):
    class ContainerWithConfig(containers.DeclarativeContainer):
        auto_load_config = False
        config = providers.Configuration(yaml_files=[yaml_config_file])

    container = ContainerWithConfig()
    assert container.config.section1.value1() is None

    container.load_config()
    assert container.config.section1.value1() == "yaml-loaded"


def test_load_config_does_not_affect_class(yaml_config_file):
    class ContainerWithConfig(containers.DeclarativeContainer):
        config = providers.Configuration(yaml_files=[yaml_config_file])

    assert ContainerWithConfig.config.section1.value1() is None
    _ = ContainerWithConfig()
    assert ContainerWithConfig.config.section1.value1() is None



"""Dict provider async mode tests."""

from dependency_injector import containers, providers
from pytest import mark


@mark.asyncio
async def test_provide():
    async def create_resource(param: str):
        return param

    class Container(containers.DeclarativeContainer):

        resources = providers.Dict(
            foo=providers.Resource(create_resource, "foo"),
            bar=providers.Resource(create_resource, "bar")
        )

    container = Container()
    resources = await container.resources()

    assert resources["foo"] == "foo"
    assert resources["bar"] == "bar"



"""FactoryAggregate provider traversal tests."""

from dependency_injector import providers


def test_traverse():
    factory1 = providers.Factory(dict)
    factory2 = providers.Factory(list)
    provider = providers.FactoryAggregate(factory1=factory1, factory2=factory2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert factory1 in all_providers
    assert factory2 in all_providers



from __future__ import annotations

from pathlib import Path
from typing import (
    Awaitable,
    TypeVar,
    Generic,
    Type,
    Callable as _Callable,
    Any,
    Tuple,
    List as _List,
    Dict as _Dict,
    Optional,
    Union,
    Coroutine as _Coroutine,
    Iterable as _Iterable,
    Iterator as _Iterator,
    AsyncIterator as _AsyncIterator,
    Generator as _Generator,
    overload,
)

try:
    import yaml
except ImportError:
    yaml = None

try:
    import pydantic
    import pydantic_settings
except ImportError:
    pydantic = None
    pydantic_settings = None

from . import resources


Injection = Any
ProviderParent = Union["Provider", Any]
T = TypeVar("T")
TT = TypeVar("TT")
P = TypeVar("P", bound="Provider")
BS = TypeVar("BS", bound="BaseSingleton")


class Provider(Generic[T]):
    def __init__(self) -> None: ...

    @overload
    def __call__(self, *args: Injection, **kwargs: Injection) -> T: ...
    @overload
    def __call__(self, *args: Injection, **kwargs: Injection) -> Awaitable[T]: ...
    def async_(self, *args: Injection, **kwargs: Injection) -> Awaitable[T]: ...

    def __deepcopy__(self, memo: Optional[_Dict[Any, Any]]) -> Provider: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    @property
    def overridden(self) -> Tuple[Provider]: ...
    @property
    def last_overriding(self) -> Optional[Provider]: ...
    def override(self, provider: Union[Provider, Any]) -> OverridingContext[P]: ...
    def reset_last_overriding(self) -> None: ...
    def reset_override(self) -> None: ...
    @property
    def overrides(self) -> Tuple[Provider]: ...
    def register_overrides(self, provider: Union[Provider, Any]) -> None: ...
    def unregister_overrides(self, provider: Union[Provider, Any]) -> None: ...
    def delegate(self) -> Provider: ...
    @property
    def provider(self) -> Provider: ...
    @property
    def provided(self) -> ProvidedInstance: ...
    def enable_async_mode(self) -> None: ...
    def disable_async_mode(self) -> None: ...
    def reset_async_mode(self) -> None: ...
    def is_async_mode_enabled(self) -> bool: ...
    def is_async_mode_disabled(self) -> bool: ...
    def is_async_mode_undefined(self) -> bool: ...
    @property
    def related(self) -> _Iterator[Provider]: ...
    def traverse(self, types: Optional[_Iterable[Type[TT]]] = None) -> _Iterator[TT]: ...
    def _copy_overridings(self, copied: Provider, memo: Optional[_Dict[Any, Any]]) -> None: ...


class Object(Provider[T]):
    def __init__(self, provides: Optional[T] = None) -> None: ...
    @property
    def provides(self) -> Optional[T]: ...
    def set_provides(self, provides: Optional[T]) -> Object: ...


class Self(Provider[T]):
    def __init__(self, container: Optional[T] = None) -> None: ...
    def set_container(self, container: T) -> None: ...
    def set_alt_names(self, alt_names: _Iterable[Any]) -> None: ...
    @property
    def alt_names(self) -> Tuple[Any]: ...


class Delegate(Provider[Provider]):
    def __init__(self, provides: Optional[Provider] = None) -> None: ...
    @property
    def provides(self) -> Optional[Provider]: ...
    def set_provides(self, provides: Optional[Provider]) -> Delegate: ...


class Aggregate(Provider[T]):
    def __init__(self, provider_dict: Optional[_Dict[Any, Provider[T]]] = None, **provider_kwargs: Provider[T]): ...
    def __getattr__(self, provider_name: Any) -> Provider[T]: ...

    @overload
    def __call__(self, provider_name: Optional[Any] = None, *args: Injection, **kwargs: Injection) -> T: ...
    @overload
    def __call__(self, provider_name: Optional[Any] = None, *args: Injection, **kwargs: Injection) -> Awaitable[T]: ...
    def async_(self, provider_name: Optional[Any] = None, *args: Injection, **kwargs: Injection) -> Awaitable[T]: ...

    @property
    def providers(self) -> _Dict[Any, Provider[T]]: ...
    def set_providers(self, provider_dict: Optional[_Dict[Any, Provider[T]]] = None, **provider_kwargs: Provider[T]) -> Aggregate[T]: ...


class Dependency(Provider[T]):
    def __init__(self, instance_of: Type[T] = object, default: Optional[Union[Provider, Any]] = None) -> None: ...
    def __getattr__(self, name: str) -> Any: ...

    @property
    def instance_of(self) -> Type[T]: ...
    def set_instance_of(self, instance_of: Type[T]) -> Dependency[T]: ...

    @property
    def default(self) -> Provider[T]: ...
    def set_default(self, default: Optional[Union[Provider, Any]]) -> Dependency[T]: ...

    @property
    def is_defined(self) -> bool: ...
    def provided_by(self, provider: Provider) -> OverridingContext[P]: ...
    @property
    def parent(self) -> Optional[ProviderParent]: ...
    @property
    def parent_name(self) -> Optional[str]: ...
    def assign_parent(self, parent: ProviderParent) -> None: ...


class ExternalDependency(Dependency[T]): ...


class DependenciesContainer(Object):
    def __init__(self, **dependencies: Provider) -> None: ...
    def __getattr__(self, name: str) -> Provider: ...
    @property
    def providers(self) -> _Dict[str, Provider]: ...
    def resolve_provider_name(self, provider: Provider) -> str: ...
    @property
    def parent(self) -> Optional[ProviderParent]: ...
    @property
    def parent_name(self) -> Optional[str]: ...
    def assign_parent(self, parent: ProviderParent) -> None: ...


class Callable(Provider[T]):
    def __init__(self, provides: Optional[Union[_Callable[..., T], str]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @property
    def provides(self) -> Optional[_Callable[..., T]]: ...
    def set_provides(self, provides: Optional[Union[_Callable[..., T], str]]) -> Callable[T]: ...
    @property
    def args(self) -> Tuple[Injection]: ...
    def add_args(self, *args: Injection) -> Callable[T]: ...
    def set_args(self, *args: Injection) -> Callable[T]: ...
    def clear_args(self) -> Callable[T]: ...
    @property
    def kwargs(self) -> _Dict[Any, Injection]: ...
    def add_kwargs(self, **kwargs: Injection) -> Callable[T]: ...
    def set_kwargs(self, **kwargs: Injection) -> Callable[T]: ...
    def clear_kwargs(self) -> Callable[T]: ...


class DelegatedCallable(Callable[T]): ...


class AbstractCallable(Callable[T]):
    def override(self, provider: Callable) -> OverridingContext[P]: ...


class CallableDelegate(Delegate):
    def __init__(self, callable: Callable) -> None: ...


class Coroutine(Callable[T]): ...


class DelegatedCoroutine(Coroutine[T]): ...


class AbstractCoroutine(Coroutine[T]):
    def override(self, provider: Coroutine) -> OverridingContext[P]: ...


class CoroutineDelegate(Delegate):
    def __init__(self, coroutine: Coroutine) -> None: ...


class ConfigurationOption(Provider[Any]):
    UNDEFINED: object
    def __init__(self, name: Tuple[str], root: Configuration) -> None: ...
    def __enter__(self) -> ConfigurationOption: ...
    def __exit__(self, *exc_info: Any) -> None: ...
    def __getattr__(self, item: str) -> ConfigurationOption: ...
    def __getitem__(self, item: Union[str, Provider]) -> ConfigurationOption: ...
    @property
    def root(self) -> Configuration: ...
    def get_name(self) -> str: ...
    def get_name_segments(self) -> Tuple[Union[str, Provider]]: ...
    def as_int(self) -> TypedConfigurationOption[int]: ...
    def as_float(self) -> TypedConfigurationOption[float]: ...
    def as_(self, callback: _Callable[..., T], *args: Injection, **kwargs: Injection) -> TypedConfigurationOption[T]: ...
    def required(self) -> ConfigurationOption: ...
    def is_required(self) -> bool: ...
    def update(self, value: Any) -> None: ...
    def from_ini(self, filepath: Union[Path, str], required: bool = False, envs_required: bool = False) -> None: ...
    def from_yaml(self, filepath: Union[Path, str], required: bool = False, loader: Optional[Any] = None, envs_required: bool = False) -> None: ...
    def from_json(self, filepath: Union[Path, str], required: bool = False, envs_required: bool = False) -> None: ...
    def from_pydantic(self, settings: PydanticSettings, required: bool = False, **kwargs: Any) -> None: ...
    def from_dict(self, options: _Dict[str, Any], required: bool = False) -> None: ...
    def from_env(self, name: str, default: Optional[Any] = None, required: bool = False, as_: Optional[_Callable[..., Any]] = None) -> None: ...
    def from_value(self, value: Any) -> None: ...


class TypedConfigurationOption(Callable[T]):
    @property
    def option(self) -> ConfigurationOption: ...


class Configuration(Object[Any]):
    DEFAULT_NAME: str = "config"
    def __init__(
            self,
            name: str = DEFAULT_NAME,
            default: Optional[Any] = None,
            *,
            strict: bool = False,
            ini_files: Optional[_Iterable[Union[Path, str]]] = None,
            yaml_files: Optional[_Iterable[Union[Path, str]]] = None,
            json_files: Optional[_Iterable[Union[Path, str]]] = None,
            pydantic_settings: Optional[_Iterable[PydanticSettings]] = None,
    ) -> None: ...
    def __enter__(self) -> Configuration : ...
    def __exit__(self, *exc_info: Any) -> None: ...
    def __getattr__(self, item: str) -> ConfigurationOption: ...
    def __getitem__(self, item: Union[str, Provider]) -> ConfigurationOption: ...

    def get_name(self) -> str: ...
    def set_name(self, name: str) -> Configuration: ...

    def get_default(self) -> _Dict[Any, Any]: ...
    def set_default(self, default: _Dict[Any, Any]): ...

    def get_strict(self) -> bool: ...
    def set_strict(self, strict: bool) -> Configuration: ...

    def get_children(self) -> _Dict[str, ConfigurationOption]: ...
    def set_children(self, children: _Dict[str, ConfigurationOption]) -> Configuration: ...

    def get_ini_files(self) -> _List[Union[Path, str]]: ...
    def set_ini_files(self, files: _Iterable[Union[Path, str]]) -> Configuration: ...

    def get_yaml_files(self) -> _List[Union[Path, str]]: ...
    def set_yaml_files(self, files: _Iterable[Union[Path, str]]) -> Configuration: ...

    def get_json_files(self) -> _List[Union[Path, str]]: ...
    def set_json_files(self, files: _Iterable[Union[Path, str]]) -> Configuration: ...

    def get_pydantic_settings(self) -> _List[PydanticSettings]: ...
    def set_pydantic_settings(self, settings: _Iterable[PydanticSettings]) -> Configuration: ...

    def load(self, required: bool = False, envs_required: bool = False) -> None: ...

    def get(self, selector: str) -> Any: ...
    def set(self, selector: str, value: Any) -> OverridingContext[P]: ...
    def reset_cache(self) -> None: ...
    def update(self, value: Any) -> None: ...
    def from_ini(self, filepath: Union[Path, str], required: bool = False, envs_required: bool = False) -> None: ...
    def from_yaml(self, filepath: Union[Path, str], required: bool = False, loader: Optional[Any] = None, envs_required: bool = False) -> None: ...
    def from_json(self, filepath: Union[Path, str], required: bool = False, envs_required: bool = False) -> None: ...
    def from_pydantic(self, settings: PydanticSettings, required: bool = False, **kwargs: Any) -> None: ...
    def from_dict(self, options: _Dict[str, Any], required: bool = False) -> None: ...
    def from_env(self, name: str, default: Optional[Any] = None, required: bool = False, as_: Optional[_Callable[..., Any]] = None) -> None: ...
    def from_value(self, value: Any) -> None: ...


class Factory(Provider[T]):
    provided_type: Optional[Type]
    def __init__(self, provides: Optional[Union[_Callable[..., T], str]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @property
    def cls(self) -> Type[T]: ...
    @property
    def provides(self) -> Optional[_Callable[..., T]]: ...
    def set_provides(self, provides: Optional[Union[_Callable[..., T], str]]) -> Factory[T]: ...
    @property
    def args(self) -> Tuple[Injection]: ...
    def add_args(self, *args: Injection) -> Factory[T]: ...
    def set_args(self, *args: Injection) -> Factory[T]: ...
    def clear_args(self) -> Factory[T]: ...
    @property
    def kwargs(self) -> _Dict[Any, Injection]: ...
    def add_kwargs(self, **kwargs: Injection) -> Factory[T]: ...
    def set_kwargs(self, **kwargs: Injection) -> Factory[T]: ...
    def clear_kwargs(self) -> Factory[T]: ...
    @property
    def attributes(self) -> _Dict[Any, Injection]: ...
    def add_attributes(self, **kwargs: Injection) -> Factory[T]: ...
    def set_attributes(self, **kwargs: Injection) -> Factory[T]: ...
    def clear_attributes(self) -> Factory[T]: ...


class DelegatedFactory(Factory[T]): ...


class AbstractFactory(Factory[T]):
    def override(self, provider: Factory) -> OverridingContext[P]: ...


class FactoryDelegate(Delegate):
    def __init__(self, factory: Factory): ...


class FactoryAggregate(Aggregate[T]):
    def __getattr__(self, provider_name: Any) -> Factory[T]: ...
    @property
    def factories(self) -> _Dict[Any, Factory[T]]: ...
    def set_factories(self, provider_dict: Optional[_Dict[Any, Factory[T]]] = None, **provider_kwargs: Factory[T]) -> FactoryAggregate[T]: ...


class BaseSingleton(Provider[T]):
    provided_type = Optional[Type]
    def __init__(self, provides: Optional[Union[_Callable[..., T], str]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @property
    def cls(self) -> Type[T]: ...
    @property
    def provides(self) -> Optional[_Callable[..., T]]: ...
    def set_provides(self, provides: Optional[Union[_Callable[..., T], str]]) -> BaseSingleton[T]: ...
    @property
    def args(self) -> Tuple[Injection]: ...
    def add_args(self, *args: Injection) -> BaseSingleton[T]: ...
    def set_args(self, *args: Injection) -> BaseSingleton[T]: ...
    def clear_args(self) -> BaseSingleton[T]: ...
    @property
    def kwargs(self) -> _Dict[Any, Injection]: ...
    def add_kwargs(self, **kwargs: Injection) -> BaseSingleton[T]: ...
    def set_kwargs(self, **kwargs: Injection) -> BaseSingleton[T]: ...
    def clear_kwargs(self) -> BaseSingleton[T]: ...
    @property
    def attributes(self) -> _Dict[Any, Injection]: ...
    def add_attributes(self, **kwargs: Injection) -> BaseSingleton[T]: ...
    def set_attributes(self, **kwargs: Injection) -> BaseSingleton[T]: ...
    def clear_attributes(self) -> BaseSingleton[T]: ...
    def reset(self) -> SingletonResetContext[BS]: ...
    def full_reset(self) -> SingletonFullResetContext[BS]: ...


class Singleton(BaseSingleton[T]): ...


class DelegatedSingleton(Singleton[T]): ...


class ThreadSafeSingleton(Singleton[T]): ...


class DelegatedThreadSafeSingleton(ThreadSafeSingleton[T]): ...


class ThreadLocalSingleton(BaseSingleton[T]): ...


class ContextLocalSingleton(BaseSingleton[T]): ...


class DelegatedThreadLocalSingleton(ThreadLocalSingleton[T]): ...


class AbstractSingleton(BaseSingleton[T]):
    def override(self, provider: BaseSingleton) -> OverridingContext[P]: ...


class SingletonDelegate(Delegate):
    def __init__(self, singleton: BaseSingleton): ...


class List(Provider[_List]):
    def __init__(self, *args: Injection): ...
    @property
    def args(self) -> Tuple[Injection]: ...
    def add_args(self, *args: Injection) -> List[T]: ...
    def set_args(self, *args: Injection) -> List[T]: ...
    def clear_args(self) -> List[T]: ...


class Dict(Provider[_Dict]):
    def __init__(self, dict_: Optional[_Dict[Any, Injection]] = None, **kwargs: Injection): ...
    @property
    def kwargs(self) -> _Dict[Any, Injection]: ...
    def add_kwargs(self, dict_: Optional[_Dict[Any, Injection]] = None, **kwargs: Injection) -> Dict: ...
    def set_kwargs(self, dict_: Optional[_Dict[Any, Injection]] = None, **kwargs: Injection) -> Dict: ...
    def clear_kwargs(self) -> Dict: ...


class Resource(Provider[T]):
    @overload
    def __init__(self, provides: Optional[Type[resources.Resource[T]]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @overload
    def __init__(self, provides: Optional[Type[resources.AsyncResource[T]]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @overload
    def __init__(self, provides: Optional[_Callable[..., _Iterator[T]]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @overload
    def __init__(self, provides: Optional[_Callable[..., _AsyncIterator[T]]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @overload
    def __init__(self, provides: Optional[_Callable[..., _Coroutine[Injection, Injection, T]]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @overload
    def __init__(self, provides: Optional[Union[_Callable[..., T], str]] = None, *args: Injection, **kwargs: Injection) -> None: ...
    @property
    def provides(self) -> Optional[_Callable[..., Any]]: ...
    def set_provides(self, provides: Optional[Any]) -> Resource[T]: ...
    @property
    def args(self) -> Tuple[Injection]: ...
    def add_args(self, *args: Injection) -> Resource[T]: ...
    def set_args(self, *args: Injection) -> Resource[T]: ...
    def clear_args(self) -> Resource[T]: ...
    @property
    def kwargs(self) -> _Dict[Any, Injection]: ...
    def add_kwargs(self, **kwargs: Injection) -> Resource[T]: ...
    def set_kwargs(self, **kwargs: Injection) -> Resource[T]: ...
    def clear_kwargs(self) -> Resource[T]: ...
    @property
    def initialized(self) -> bool: ...
    def init(self) -> Optional[Awaitable[T]]: ...
    def shutdown(self) -> Optional[Awaitable]: ...


class Container(Provider[T]):
    def __init__(self, container_cls: Type[T], container: Optional[T] = None, **overriding_providers: Union[Provider, Any]) -> None: ...
    def __getattr__(self, name: str) -> Provider: ...
    @property
    def container(self) -> T: ...
    def resolve_provider_name(self, provider: Provider) -> str: ...
    @property
    def parent(self) -> Optional[ProviderParent]: ...
    @property
    def parent_name(self) -> Optional[str]: ...
    def assign_parent(self, parent: ProviderParent) -> None: ...


class Selector(Provider[Any]):
    def __init__(self, selector: Optional[_Callable[..., Any]] = None, **providers: Provider): ...
    def __getattr__(self, name: str) -> Provider: ...

    @property
    def selector(self) -> Optional[_Callable[..., Any]]: ...
    def set_selector(self, selector: Optional[_Callable[..., Any]]) -> Selector: ...

    @property
    def providers(self) -> _Dict[str, Provider]: ...
    def set_providers(self, **providers: Provider) -> Selector: ...


class ProvidedInstanceFluentInterface:
    def __getattr__(self, item: Any) -> AttributeGetter: ...
    def __getitem__(self, item: Any) -> ItemGetter: ...
    def call(self, *args: Injection, **kwargs: Injection) -> MethodCaller: ...
    @property
    def provides(self) -> Optional[Provider]: ...
    def set_provides(self, provides: Optional[Provider]) -> ProvidedInstanceFluentInterface: ...


class ProvidedInstance(Provider, ProvidedInstanceFluentInterface):
    def __init__(self, provides: Optional[Provider] = None) -> None: ...


class AttributeGetter(Provider, ProvidedInstanceFluentInterface):
    def __init__(self, provides: Optional[Provider] = None, name: Optional[str] = None) -> None: ...
    @property
    def name(self) -> Optional[str]: ...
    def set_name(self, name: Optional[str]) -> ProvidedInstanceFluentInterface: ...


class ItemGetter(Provider, ProvidedInstanceFluentInterface):
    def __init__(self, provides: Optional[Provider] = None, name: Optional[str] = None) -> None: ...
    @property
    def name(self) -> Optional[str]: ...
    def set_name(self, name: Optional[str]) -> ProvidedInstanceFluentInterface: ...


class MethodCaller(Provider, ProvidedInstanceFluentInterface):
    def __init__(self, provides: Optional[Provider] = None, *args: Injection, **kwargs: Injection) -> None: ...


class OverridingContext(Generic[T]):
    def __init__(self, overridden: Provider, overriding: Provider): ...
    def __enter__(self) -> T: ...
    def __exit__(self, *_: Any) -> None:
        pass
    ...


class BaseSingletonResetContext(Generic[T]):
    def __init__(self, provider: T): ...
    def __enter__(self) -> T: ...
    def __exit__(self, *_: Any) -> None: ...


class SingletonResetContext(BaseSingletonResetContext):
    ...


class SingletonFullResetContext(BaseSingletonResetContext):
    ...


CHILD_PROVIDERS: Tuple[Provider]


def is_provider(instance: Any) -> bool: ...


def ensure_is_provider(instance: Any) -> Provider: ...


def is_delegated(instance: Any) -> bool: ...


def represent_provider(provider: Provider, provides: Any) -> str: ...


def deepcopy(instance: Any, memo: Optional[_Dict[Any, Any]] = None): Any: ...


def merge_dicts(dict1: _Dict[Any, Any], dict2: _Dict[Any, Any]) -> _Dict[Any, Any]: ...


def traverse(*providers: Provider, types: Optional[_Iterable[Type]]=None) -> _Iterator[Provider]: ...


if yaml:
    class YamlLoader(yaml.SafeLoader): ...
else:
    class YamlLoader: ...

if pydantic:
    PydanticSettings = pydantic_settings.BaseSettings
else:
    PydanticSettings = Any



"""Container provider traversal tests."""

from dependency_injector import containers, providers


def test_traverse():
    class Container(containers.DeclarativeContainer):
        provider1 = providers.Callable(list)
        provider2 = providers.Callable(dict)

    provider = providers.Container(Container)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert {list, dict} == {provider.provides for provider in all_providers}


def test_traverse_overridden():
    class Container1(containers.DeclarativeContainer):
        provider1 = providers.Callable(list)
        provider2 = providers.Callable(dict)

    class Container2(containers.DeclarativeContainer):
        provider1 = providers.Callable(tuple)
        provider2 = providers.Callable(str)

    container2 = Container2()

    provider = providers.Container(Container1)
    provider.override(container2)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 5
    assert {list, dict, tuple, str} == {
        provider.provides
        for provider in all_providers
        if isinstance(provider, providers.Callable)
    }
    assert provider.last_overriding in all_providers
    assert provider.last_overriding() is container2



"""Container typing in runtime tests."""

from dependency_injector import containers


def test_types_declarative():
    container: containers.Container = containers.DeclarativeContainer()
    assert isinstance(container, containers.Container)


def test_types_dynamic():
    container: containers.Container = containers.DynamicContainer()
    assert isinstance(container, containers.Container)



"""Tests for configuration provider linking."""


from dependency_injector import containers, providers


class Core(containers.DeclarativeContainer):
    config = providers.Configuration("core")
    value_getter = providers.Callable(lambda _: _, config.value)


class Services(containers.DeclarativeContainer):
    config = providers.Configuration("services")
    value_getter = providers.Callable(lambda _: _, config.value)


def test():
    root_config = providers.Configuration("main")
    core = Core(config=root_config.core)
    services = Services(config=root_config.services)

    root_config.override(
        {
            "core": {
                "value": "core",
            },
            "services": {
                "value": "services",
            },
        },
    )

    assert core.config() == {"value": "core"}
    assert core.config.value() == "core"
    assert core.value_getter() == "core"

    assert services.config() == {"value": "services"}
    assert services.config.value() == "services"
    assert services.value_getter() == "services"


def test_double_override():
    root_config = providers.Configuration("main")
    core = Core(config=root_config.core)
    services = Services(config=root_config.services)

    root_config.override(
        {
            "core": {
                "value": "core1",
            },
            "services": {
                "value": "services1",
            },
        },
    )
    root_config.override(
        {
            "core": {
                "value": "core2",
            },
            "services": {
                "value": "services2",
            },
        },
    )

    assert core.config() == {"value": "core2"}
    assert core.config.value() == "core2"
    assert core.value_getter() == "core2"

    assert services.config() == {"value": "services2"}
    assert services.config.value() == "services2"
    assert services.value_getter() == "services2"


def test_reset_overriding_cache():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/428
    class Core(containers.DeclarativeContainer):
        config = providers.Configuration()

        greetings = providers.Factory(str, config.greeting)

    class Application(containers.DeclarativeContainer):
        config = providers.Configuration()

        core = providers.Container(
            Core,
            config=config,
        )

        greetings = providers.Factory(str, config.greeting)

    container = Application()

    container.config.set("greeting", "Hello World")
    assert container.greetings() == "Hello World"
    assert container.core.greetings() == "Hello World"

    container.config.set("greeting", "Hello Bob")
    assert container.greetings() == "Hello Bob"
    assert container.core.greetings() == "Hello Bob"


def test_reset_overriding_cache_for_option():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/428
    class Core(containers.DeclarativeContainer):
        config = providers.Configuration()

        greetings = providers.Factory(str, config.greeting)

    class Application(containers.DeclarativeContainer):
        config = providers.Configuration()

        core = providers.Container(
            Core,
            config=config.option,
        )

        greetings = providers.Factory(str, config.option.greeting)

    container = Application()

    container.config.set("option.greeting", "Hello World")
    assert container.greetings() == "Hello World"
    assert container.core.greetings() == "Hello World"

    container.config.set("option.greeting", "Hello Bob")
    assert container.greetings() == "Hello Bob"
    assert container.core.greetings() == "Hello Bob"



from dependency_injector import providers


# Test 1: to check .provided attribute
provider1: providers.Provider[int] = providers.Object(1)
provided: providers.ProvidedInstance = provider1.provided

# Test 2: to check async mode API
provider2: providers.Provider = providers.Provider()
provider2.enable_async_mode()
provider2.disable_async_mode()
provider2.reset_async_mode()
r1: bool = provider2.is_async_mode_enabled()
r2: bool = provider2.is_async_mode_disabled()
r3: bool = provider2.is_async_mode_undefined()



"""Container tests for building containers from configuration files."""

import os
import sqlite3

from dependency_injector import containers

from samples.schema.services import UserService, AuthService, PhotoService


SAMPLES_DIR = os.path.abspath(
    os.path.sep.join((
        os.path.dirname(__file__),
        "../samples/",
    )),
)


def test_single_container_schema(container: containers.DynamicContainer):
    container.from_yaml_schema(f"{SAMPLES_DIR}/schema/container-single.yml")
    container.config.from_dict(
        {
            "database": {
                "dsn": ":memory:",
            },
            "aws": {
                "access_key_id": "KEY",
                "secret_access_key": "SECRET",
            },
            "auth": {
                "token_ttl": 3600,
            },
        },
    )

    # User service
    user_service1 = container.user_service()
    user_service2 = container.user_service()
    assert isinstance(user_service1, UserService)
    assert isinstance(user_service2, UserService)
    assert user_service1 is not user_service2

    assert isinstance(user_service1.db, sqlite3.Connection)
    assert isinstance(user_service2.db, sqlite3.Connection)
    assert user_service1.db is user_service2.db

    # Auth service
    auth_service1 = container.auth_service()
    auth_service2 = container.auth_service()
    assert isinstance(auth_service1, AuthService)
    assert isinstance(auth_service2, AuthService)
    assert auth_service1 is not auth_service2

    assert isinstance(auth_service1.db, sqlite3.Connection)
    assert isinstance(auth_service2.db, sqlite3.Connection)
    assert auth_service1.db is auth_service2.db
    assert auth_service1.db is container.database_client()
    assert auth_service2.db is container.database_client()

    assert auth_service1.token_ttl == 3600
    assert auth_service2.token_ttl == 3600

    # Photo service
    photo_service1 = container.photo_service()
    photo_service2 = container.photo_service()
    assert isinstance(photo_service1, PhotoService)
    assert isinstance(photo_service2, PhotoService)
    assert photo_service1 is not photo_service2

    assert isinstance(photo_service1.db, sqlite3.Connection)
    assert isinstance(photo_service2.db, sqlite3.Connection)
    assert photo_service1.db is photo_service2.db
    assert photo_service1.db is container.database_client()
    assert photo_service2.db is container.database_client()

    assert photo_service1.s3 is photo_service2.s3
    assert photo_service1.s3 is container.s3_client()
    assert photo_service2.s3 is container.s3_client()


def test_multiple_containers_schema(container: containers.DynamicContainer):
    container.from_yaml_schema(f"{SAMPLES_DIR}/schema/container-multiple.yml")
    container.core.config.from_dict(
        {
            "database": {
                "dsn": ":memory:",
            },
            "aws": {
                "access_key_id": "KEY",
                "secret_access_key": "SECRET",
            },
            "auth": {
                "token_ttl": 3600,
            },
        },
    )

    # User service
    user_service1 = container.services.user()
    user_service2 = container.services.user()
    assert isinstance(user_service1, UserService)
    assert isinstance(user_service2, UserService)
    assert user_service1 is not user_service2

    assert isinstance(user_service1.db, sqlite3.Connection)
    assert isinstance(user_service2.db, sqlite3.Connection)
    assert user_service1.db is user_service2.db

    # Auth service
    auth_service1 = container.services.auth()
    auth_service2 = container.services.auth()
    assert isinstance(auth_service1, AuthService)
    assert isinstance(auth_service2, AuthService)
    assert auth_service1 is not auth_service2

    assert isinstance(auth_service1.db, sqlite3.Connection)
    assert isinstance(auth_service2.db, sqlite3.Connection)
    assert auth_service1.db is auth_service2.db
    assert auth_service1.db is container.gateways.database_client()
    assert auth_service2.db is container.gateways.database_client()

    assert auth_service1.token_ttl == 3600
    assert auth_service2.token_ttl == 3600

    # Photo service
    photo_service1 = container.services.photo()
    photo_service2 = container.services.photo()
    assert isinstance(photo_service1, PhotoService)
    assert isinstance(photo_service2, PhotoService)
    assert photo_service1 is not photo_service2

    assert isinstance(photo_service1.db, sqlite3.Connection)
    assert isinstance(photo_service2.db, sqlite3.Connection)
    assert photo_service1.db is photo_service2.db
    assert photo_service1.db is container.gateways.database_client()
    assert photo_service2.db is container.gateways.database_client()

    assert photo_service1.s3 is photo_service2.s3
    assert photo_service1.s3 is container.gateways.s3_client()
    assert photo_service2.s3 is container.gateways.s3_client()


def test_multiple_reordered_containers_schema(container: containers.DynamicContainer):
    container.from_yaml_schema(f"{SAMPLES_DIR}/schema/container-multiple-reordered.yml")
    container.core.config.from_dict(
        {
            "database": {
                "dsn": ":memory:",
            },
            "aws": {
                "access_key_id": "KEY",
                "secret_access_key": "SECRET",
            },
            "auth": {
                "token_ttl": 3600,
            },
        },
    )

    # User service
    user_service1 = container.services.user()
    user_service2 = container.services.user()
    assert isinstance(user_service1, UserService)
    assert isinstance(user_service2, UserService)
    assert user_service1 is not user_service2

    assert isinstance(user_service1.db, sqlite3.Connection)
    assert isinstance(user_service2.db, sqlite3.Connection)
    assert user_service1.db is user_service2.db

    # Auth service
    auth_service1 = container.services.auth()
    auth_service2 = container.services.auth()
    assert isinstance(auth_service1, AuthService)
    assert isinstance(auth_service2, AuthService)
    assert auth_service1 is not auth_service2

    assert isinstance(auth_service1.db, sqlite3.Connection)
    assert isinstance(auth_service2.db, sqlite3.Connection)
    assert auth_service1.db is auth_service2.db
    assert auth_service1.db is container.gateways.database_client()
    assert auth_service2.db is container.gateways.database_client()

    assert auth_service1.token_ttl == 3600
    assert auth_service2.token_ttl == 3600

    # Photo service
    photo_service1 = container.services.photo()
    photo_service2 = container.services.photo()
    assert isinstance(photo_service1, PhotoService)
    assert isinstance(photo_service2, PhotoService)
    assert photo_service1 is not photo_service2

    assert isinstance(photo_service1.db, sqlite3.Connection)
    assert isinstance(photo_service2.db, sqlite3.Connection)
    assert photo_service1.db is photo_service2.db
    assert photo_service1.db is container.gateways.database_client()
    assert photo_service2.db is container.gateways.database_client()

    assert photo_service1.s3 is photo_service2.s3
    assert photo_service1.s3 is container.gateways.s3_client()
    assert photo_service2.s3 is container.gateways.s3_client()


def test_multiple_containers_with_inline_providers_schema(container: containers.DynamicContainer):
    container.from_yaml_schema(f"{SAMPLES_DIR}/schema/container-multiple-inline.yml")
    container.core.config.from_dict(
        {
            "database": {
                "dsn": ":memory:",
            },
            "aws": {
                "access_key_id": "KEY",
                "secret_access_key": "SECRET",
            },
            "auth": {
                "token_ttl": 3600,
            },
        },
    )

    # User service
    user_service1 = container.services.user()
    user_service2 = container.services.user()
    assert isinstance(user_service1, UserService)
    assert isinstance(user_service2, UserService)
    assert user_service1 is not user_service2

    assert isinstance(user_service1.db, sqlite3.Connection)
    assert isinstance(user_service2.db, sqlite3.Connection)
    assert user_service1.db is user_service2.db

    # Auth service
    auth_service1 = container.services.auth()
    auth_service2 = container.services.auth()
    assert isinstance(auth_service1, AuthService)
    assert isinstance(auth_service2, AuthService)
    assert auth_service1 is not auth_service2

    assert isinstance(auth_service1.db, sqlite3.Connection)
    assert isinstance(auth_service2.db, sqlite3.Connection)
    assert auth_service1.db is auth_service2.db
    assert auth_service1.db is container.gateways.database_client()
    assert auth_service2.db is container.gateways.database_client()

    assert auth_service1.token_ttl == 3600
    assert auth_service2.token_ttl == 3600

    # Photo service
    photo_service1 = container.services.photo()
    photo_service2 = container.services.photo()
    assert isinstance(photo_service1, PhotoService)
    assert isinstance(photo_service2, PhotoService)
    assert photo_service1 is not photo_service2

    assert isinstance(photo_service1.db, sqlite3.Connection)
    assert isinstance(photo_service2.db, sqlite3.Connection)
    assert photo_service1.db is photo_service2.db
    assert photo_service1.db is container.gateways.database_client()
    assert photo_service2.db is container.gateways.database_client()

    assert photo_service1.s3 is photo_service2.s3
    assert photo_service1.s3 is container.gateways.s3_client()
    assert photo_service2.s3 is container.gateways.s3_client()


def test_schema_with_boto3_session(container: containers.DynamicContainer):
    container.from_yaml_schema(f"{SAMPLES_DIR}/schema/container-boto3-session.yml")
    container.config.from_dict(
        {
            "aws_access_key_id": "key",
            "aws_secret_access_key": "secret",
            "aws_session_token": "token",
            "aws_region_name": "us-east-1",
        },
    )

    assert container.s3_client().__class__.__name__ == "S3"
    assert container.sqs_client().__class__.__name__ == "SQS"



"""Fixtures module."""

import json
import os

from dependency_injector import providers
from pytest import fixture


@fixture
def config_type():
    return "default"


@fixture
def config(config_type):
    if config_type == "strict":
        return providers.Configuration(strict=True)
    elif config_type == "default":
        return providers.Configuration()
    else:
        raise ValueError("Undefined config type \"{0}\"".format(config_type))


@fixture
def ini_config_file_1(tmp_path):
    config_file = str(tmp_path / "config_1.ini")
    with open(config_file, "w") as file:
        file.write(
            "[section1]\n"
            "value1=1\n"
            "\n"
            "[section2]\n"
            "value2=2\n"
        )
    return config_file


@fixture
def ini_config_file_2(tmp_path):
    config_file = str(tmp_path / "config_2.ini")
    with open(config_file, "w") as file:
        file.write(
            "[section1]\n"
            "value1=11\n"
            "value11=11\n"
            "[section3]\n"
            "value3=3\n"
        )
    return config_file


@fixture
def ini_config_file_3(tmp_path):
    config_file = str(tmp_path / "config_3.ini")
    with open(config_file, "w") as file:
        file.write(
            "[section1]\n"
            "value1=${CONFIG_TEST_ENV}\n"
            "value2=${CONFIG_TEST_PATH}/path\n"
        )
    return config_file


@fixture
def yaml_config_file_1(tmp_path):
    config_file = str(tmp_path / "config_1.yml")
    with open(config_file, "w") as file:
        file.write(
            "section1:\n"
            "  value1: 1\n"
            "\n"
            "section2:\n"
            "  value2: 2\n"
        )
    return config_file


@fixture
def yaml_config_file_2(tmp_path):
    config_file = str(tmp_path / "config_2.yml")
    with open(config_file, "w") as file:
        file.write(
            "section1:\n"
            "  value1: 11\n"
            "  value11: 11\n"
            "section3:\n"
            "  value3: 3\n"
        )
    return config_file


@fixture
def yaml_config_file_3(tmp_path):
    config_file = str(tmp_path / "config_3.yml")
    with open(config_file, "w") as file:
        file.write(
            "section1:\n"
            "  value1: ${CONFIG_TEST_ENV}\n"
            "  value2: ${CONFIG_TEST_PATH}/path\n"
        )
    return config_file


@fixture
def json_config_file_1(tmp_path):
    config_file = str(tmp_path / "config_1.json")
    with open(config_file, "w") as file:
        file.write(
            json.dumps(
                {
                    "section1": {
                        "value1": 1,
                    },
                    "section2": {
                        "value2": 2,
                    },
                },
            ),
        )
    return config_file


@fixture
def json_config_file_2(tmp_path):
    config_file = str(tmp_path / "config_2.json")
    with open(config_file, "w") as file:
        file.write(
            json.dumps(
                {
                    "section1": {
                        "value1": 11,
                        "value11": 11,
                    },
                    "section3": {
                        "value3": 3,
                    },
                },
            ),
        )
    return config_file


@fixture
def json_config_file_3(tmp_path):
    config_file = str(tmp_path / "config_3.json")
    with open(config_file, "w") as file:
        file.write(
            json.dumps(
                {
                    "section1": {
                        "value1": "${CONFIG_TEST_ENV}",
                        "value2": "${CONFIG_TEST_PATH}/path",
                    },
                },
            ),
        )
    return config_file


@fixture(autouse=True)
def environment_variables():
    os.environ["CONFIG_TEST_ENV"] = "test-value"
    os.environ["CONFIG_TEST_PATH"] = "test-path"
    os.environ["DEFINED"] = "defined"
    os.environ["EMPTY"] = ""
    os.environ["CONFIG_INT"] = "42"
    yield
    os.environ.pop("CONFIG_TEST_ENV", None)
    os.environ.pop("CONFIG_TEST_PATH", None)
    os.environ.pop("DEFINED", None)
    os.environ.pop("EMPTY", None)
    os.environ.pop("CONFIG_INT", None)



"""Test module for wiring."""

from decimal import Decimal
from typing import Callable

from dependency_injector import providers
from dependency_injector.wiring import inject, Provide, Provider

from .container import Container, SubContainer
from .service import Service


service: Service = Provide[Container.service]
service_provider: Callable[..., Service] = Provider[Container.service]
undefined: Callable = Provide[providers.Provider()]


class TestClass:

    service: Service = Provide[Container.service]
    service_provider: Callable[..., Service] = Provider[Container.service]
    undefined: Callable = Provide[providers.Provider()]

    @inject
    def __init__(self, service: Service = Provide[Container.service]):
        self.service = service

    @inject
    def method(self, service: Service = Provide[Container.service]):
        return service

    @classmethod
    @inject
    def class_method(cls, service: Service = Provide[Container.service]):
        return service

    @staticmethod
    @inject
    def static_method(service: Service = Provide[Container.service]):
        return service


@inject
def test_function(service: Service = Provide[Container.service]):
    return service


@inject
def test_function_provider(service_provider: Callable[..., Service] = Provider[Container.service]):
    service = service_provider()
    return service


@inject
def test_config_value(
        value_int: int = Provide[Container.config.a.b.c.as_int()],
        value_float: float = Provide[Container.config.a.b.c.as_float()],
        value_str: str = Provide[Container.config.a.b.c.as_(str)],
        value_decimal: Decimal = Provide[Container.config.a.b.c.as_(Decimal)],
        value_required: str = Provide[Container.config.a.b.c.required()],
        value_required_int: int = Provide[Container.config.a.b.c.required().as_int()],
        value_required_float: float = Provide[Container.config.a.b.c.required().as_float()],
        value_required_str: str = Provide[Container.config.a.b.c.required().as_(str)],
        value_required_decimal: str = Provide[Container.config.a.b.c.required().as_(Decimal)],
):
    return (
        value_int,
        value_float,
        value_str,
        value_decimal,
        value_required,
        value_required_int,
        value_required_float,
        value_required_str,
        value_required_decimal,
    )


@inject
def test_config_value_required_undefined(
        value_required: int = Provide[Container.config.a.b.c.required()],
):
    return value_required


@inject
def test_provide_provider(service_provider: Callable[..., Service] = Provide[Container.service.provider]):
    service = service_provider()
    return service


@inject
def test_provider_provider(service_provider: Callable[..., Service] = Provider[Container.service.provider]):
    service = service_provider()
    return service


@inject
def test_provided_instance(some_value: int = Provide[Container.service.provided.foo["bar"].call()]):
    return some_value


@inject
def test_subcontainer_provider(some_value: int = Provide[Container.sub.int_object]):
    return some_value


@inject
def test_config_invariant(some_value: int = Provide[Container.config.option[Container.config.switch]]):
    return some_value


@inject
def test_provide_from_different_containers(
        service: Service = Provide[Container.service],
        some_value: int = Provide[SubContainer.int_object],
):
    return service, some_value


class ClassDecorator:
    def __init__(self, fn):
        self._fn = fn

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


@ClassDecorator
@inject
def test_class_decorator(service: Service = Provide[Container.service]):
    return service


def test_container(container: Container = Provide[Container]):
    return container.service()



"""Wiring optimizations module."""

import asyncio
import collections.abc
import functools
import inspect
import types

from . import providers
from .wiring import _Marker, PatchedCallable

from .providers cimport Provider


def _get_sync_patched(fn, patched: PatchedCallable):
    @functools.wraps(fn)
    def _patched(*args, **kwargs):
        cdef object result
        cdef dict to_inject
        cdef object arg_key
        cdef Provider provider

        to_inject = kwargs.copy()
        for arg_key, provider in patched.injections.items():
            if arg_key not in kwargs or isinstance(kwargs[arg_key], _Marker):
                to_inject[arg_key] = provider()

        result = fn(*args, **to_inject)

        if patched.closing:
            for arg_key, provider in patched.closing.items():
                if arg_key in kwargs and not isinstance(kwargs[arg_key], _Marker):
                    continue
                if not isinstance(provider, providers.Resource):
                    continue
                provider.shutdown()

        return result
    return _patched


async def _async_inject(object fn, tuple args, dict kwargs, dict injections, dict closings):
    cdef object result
    cdef dict to_inject
    cdef list to_inject_await = []
    cdef list to_close_await = []
    cdef object arg_key
    cdef Provider provider

    to_inject = kwargs.copy()
    for arg_key, provider in injections.items():
        if arg_key not in kwargs or isinstance(kwargs[arg_key], _Marker):
            provide = provider()
            if provider.is_async_mode_enabled():
                to_inject_await.append((arg_key, provide))
            elif _isawaitable(provide):
                to_inject_await.append((arg_key, provide))
            else:
                to_inject[arg_key] = provide

    if to_inject_await:
        async_to_inject = await asyncio.gather(*(provide for _, provide in to_inject_await))
        for provide, (injection, _) in zip(async_to_inject, to_inject_await):
            to_inject[injection] = provide

    result = await fn(*args, **to_inject)

    if closings:
        for arg_key, provider in closings.items():
            if arg_key in kwargs and isinstance(kwargs[arg_key], _Marker):
                continue
            if not isinstance(provider, providers.Resource):
                continue
            shutdown = provider.shutdown()
            if _isawaitable(shutdown):
                to_close_await.append(shutdown)

        await asyncio.gather(*to_close_await)

    return result


cdef bint _isawaitable(object instance):
    """Return true if object can be passed to an ``await`` expression."""
    return (isinstance(instance, types.CoroutineType) or
            isinstance(instance, types.GeneratorType) and
            bool(instance.gi_code.co_flags & inspect.CO_ITERABLE_COROUTINE) or
            isinstance(instance, collections.abc.Awaitable))



from typing import List, Iterator, Generator, AsyncIterator, AsyncGenerator, Optional

from dependency_injector import providers, resources


# Test 1: to check the return type with function
def init1() -> List[int]:
    return []


provider1 = providers.Resource(init1)
var1: List[int] = provider1()


# Test 2: to check the return type with iterator
def init2() -> Iterator[List[int]]:
    yield []


provider2 = providers.Resource(init2)
var2: List[int] = provider2()


# Test 3: to check the return type with generator
def init3() -> Generator[List[int], None, None]:
    yield []


provider3 = providers.Resource(init3)
var3: List[int] = provider3()


# Test 4: to check the return type with resource subclass
class MyResource4(resources.Resource[List[int]]):
    def init(self, *args, **kwargs) -> List[int]:
        return []

    def shutdown(self, resource: Optional[List[int]]) -> None:
        ...


provider4 = providers.Resource(MyResource4)
var4: List[int] = provider4()


# Test 5: to check the return type with async function
async def init5() -> List[int]:
    return []


provider5 = providers.Resource(init5)


async def _provide5() -> None:
    var1: List[int] = await provider5()  # type: ignore
    var2: List[int] = await provider5.async_()


# Test 6: to check the return type with async iterator
async def init6() -> AsyncIterator[List[int]]:
    yield []


provider6 = providers.Resource(init6)


async def _provide6() -> None:
    var1: List[int] = await provider6()  # type: ignore
    var2: List[int] = await provider6.async_()


# Test 7: to check the return type with async generator
async def init7() -> AsyncGenerator[List[int], None]:
    yield []


provider7 = providers.Resource(init7)


async def _provide7() -> None:
    var1: List[int] = await provider7()  # type: ignore
    var2: List[int] = await provider7.async_()


# Test 8: to check the return type with async resource subclass
class MyResource8(resources.AsyncResource[List[int]]):
    async def init(self, *args, **kwargs) -> List[int]:
        return []

    async def shutdown(self, resource: Optional[List[int]]) -> None:
        ...


provider8 = providers.Resource(MyResource8)


async def _provide8() -> None:
    var1: List[int] = await provider8()  # type: ignore
    var2: List[int] = await provider8.async_()


# Test 9: to check string imports
provider9: providers.Resource[dict] = providers.Resource("builtins.dict")
provider9.set_provides("builtins.dict")



"""Container API tests for building container from schema."""

import contextlib
import json
import pathlib
import re

import yaml
from dependency_injector import containers, providers, errors
from pytest import raises


def test_from_schema(container: containers.DynamicContainer):
    container.from_schema(
        {
            "version": "1",
            "container": {
                "provider1": {
                    "provider": "Factory",
                    "provides": "list",
                    "args": [1, 2, 3],
                },
                "provider2": {
                    "provider": "Factory",
                    "provides": "dict",
                    "kwargs": {
                        "one": "container.provider1",
                        "two": 2,
                    },
                },
            },
        },
    )

    assert isinstance(container.provider1, providers.Factory)
    assert container.provider1.provides is list
    assert container.provider1.args == (1, 2, 3)

    assert isinstance(container.provider2, providers.Factory)
    assert container.provider2.provides is dict
    assert container.provider2.kwargs == {"one": container.provider1, "two": 2}


def test_from_yaml_schema(container: containers.DynamicContainer, tmp_path: pathlib.Path):
    schema_path = tmp_path / "schema.yml"
    with open(schema_path, "w") as file:
        file.write("""
        version: "1"
        container:
          provider1:
            provider: Factory
            provides: list
            args:
              - 1
              - 2
              - 3
          provider2:
            provider: Factory
            provides: dict
            kwargs:
              one: container.provider1
              two: 2
        """)
    container.from_yaml_schema(schema_path)

    assert isinstance(container.provider1, providers.Factory)
    assert container.provider1.provides == list
    assert container.provider1.args == (1, 2, 3)

    assert isinstance(container.provider2, providers.Factory)
    assert container.provider2.provides is dict
    assert container.provider2.kwargs == {"one": container.provider1, "two": 2}


def test_from_yaml_schema_with_loader(container: containers.DynamicContainer, tmp_path: pathlib.Path):
    schema_path = tmp_path / "schema.yml"
    with open(schema_path, "w") as file:
        file.write("""
        version: "1"
        container:
          provider:
            provider: Factory
            provides: list
            args: [1, 2, 3]
        """)
    container.from_yaml_schema(schema_path, loader=yaml.Loader)

    assert isinstance(container.provider, providers.Factory)
    assert container.provider.provides is list
    assert container.provider.args == (1, 2, 3)


def test_from_yaml_schema_no_yaml_installed(container: containers.DynamicContainer):
    @contextlib.contextmanager
    def no_yaml_module():
        containers.yaml = None
        yield
        containers.yaml = yaml

    error_message = re.escape(
        "Unable to load yaml schema - PyYAML is not installed. "
        "Install PyYAML or install Dependency Injector with yaml extras: "
        "\"pip install dependency-injector[yaml]\""
    )

    with no_yaml_module():
        with raises(errors.Error, match=error_message):
            container.from_yaml_schema("./no-yaml-installed.yml")


def test_from_json_schema(container: containers.DynamicContainer, tmp_path: pathlib.Path):
    schema_path = tmp_path / "schema.json"
    with open(schema_path, "w") as file:
        file.write(
            json.dumps(
                {
                    "version": "1",
                    "container": {
                        "provider1": {
                            "provider": "Factory",
                            "provides": "list",
                            "args": [1, 2, 3],
                        },
                        "provider2": {
                            "provider": "Factory",
                            "provides": "dict",
                            "kwargs": {
                                "one": "container.provider1",
                                "two": 2,
                            },
                        },
                    },
                },
                indent=4,
            ),
        )
    container.from_json_schema(schema_path)

    assert isinstance(container.provider1, providers.Factory)
    assert container.provider1.provides is list
    assert container.provider1.args == (1, 2, 3)

    assert isinstance(container.provider2, providers.Factory)
    assert container.provider2.provides is dict
    assert container.provider2.kwargs == {"one": container.provider1, "two": 2}



"""Selector provider tests."""

import functools
import itertools
import sys

from dependency_injector import providers, errors
from pytest import fixture, mark, raises


@fixture
def switch():
    return providers.Configuration()


@fixture
def one():
    return providers.Object(1)


@fixture
def two():
    return providers.Object(2)


@fixture
def selector_type():
    return "default"


@fixture
def selector(selector_type, switch, one, two):
    if selector_type == "default":
        return providers.Selector(switch, one=one, two=two)
    elif selector_type == "empty":
        return providers.Selector()
    elif selector_type == "sys-streams":
        return providers.Selector(
            switch,
            stdin=providers.Object(sys.stdin),
            stdout=providers.Object(sys.stdout),
            stderr=providers.Object(sys.stderr),
        )
    else:
        raise ValueError("Unknown selector type \"{0}\"".format(selector_type))


def test_is_provider(selector):
    assert providers.is_provider(selector) is True


@mark.parametrize("selector_type", ["empty"])
def test_init_optional(selector, switch, one, two):
    selector.set_selector(switch)
    selector.set_providers(one=one, two=two)

    assert selector.providers == {"one": one, "two": two}
    with switch.override("one"):
        assert selector() == one()
    with switch.override("two"):
        assert selector() == two()


def test_set_selector_returns_self(selector, switch):
    assert selector.set_selector(switch) is selector


def test_set_providers_returns_self(selector, one):
    assert selector.set_providers(one=one) is selector


def test_provided_instance_provider(selector):
    assert isinstance(selector.provided, providers.ProvidedInstance)


def test_call(selector, switch):
    with switch.override("one"):
        assert selector() == 1

    with switch.override("two"):
        assert selector() == 2


def test_call_undefined_provider(selector, switch):
    with switch.override("three"):
        with raises(errors.Error):
            selector()


def test_call_selector_is_none(selector, switch):
    with switch.override(None):
        with raises(errors.Error):
            selector()


@mark.parametrize("selector_type", ["empty"])
def test_call_any_callable(selector):
    selector.set_selector(functools.partial(next, itertools.cycle(["one", "two"])))
    selector.set_providers(
        one=providers.Object(1),
        two=providers.Object(2),
    )

    assert selector() == 1
    assert selector() == 2
    assert selector() == 1
    assert selector() == 2


@mark.parametrize("selector_type", ["empty"])
def test_call_with_context_args(selector, switch):
    selector.set_selector(switch)
    selector.set_providers(one=providers.Callable(lambda *args, **kwargs: (args, kwargs)))

    with switch.override("one"):
        args, kwargs = selector(1, 2, three=3, four=4)

    assert args == (1, 2)
    assert kwargs == {"three": 3, "four": 4}


def test_getattr(selector, one, two):
    assert selector.one is one
    assert selector.two is two


def test_getattr_attribute_error(selector):
    with raises(AttributeError):
        _ = selector.provider_three


def test_call_overridden(selector, switch):
    overriding_provider1 = providers.Selector(switch, one=providers.Object(2))
    overriding_provider2 = providers.Selector(switch, one=providers.Object(3))

    selector.override(overriding_provider1)
    selector.override(overriding_provider2)

    with switch.override("one"):
        assert selector() == 3


def test_providers_attribute(selector, one, two):
    assert selector.providers == {"one": one, "two": two}


def test_deepcopy(selector):
    provider_copy = providers.deepcopy(selector)

    assert provider_copy is not selector
    assert isinstance(selector, providers.Selector)

    assert provider_copy.selector is not selector.selector
    assert isinstance(provider_copy.selector, providers.Configuration)

    assert provider_copy.one is not selector.one
    assert isinstance(provider_copy.one, providers.Object)
    assert provider_copy.one.provides == 1

    assert provider_copy.two is not selector.two
    assert isinstance(provider_copy.two, providers.Object)
    assert provider_copy.two.provides == 2


def test_deepcopy_from_memo(selector):
    provider_copy = providers.deepcopy(
        selector,
        memo={id(selector): selector},
    )
    assert provider_copy is selector


def test_deepcopy_overridden(selector):
    object_provider = providers.Object(object())

    selector.override(object_provider)

    provider_copy = providers.deepcopy(selector)
    object_provider_copy = provider_copy.overridden[0]

    assert selector is not provider_copy
    assert isinstance(selector, providers.Selector)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


@mark.parametrize("selector_type", ["sys-streams"])
def test_deepcopy_with_sys_streams(selector, switch):
    provider_copy = providers.deepcopy(selector)

    assert selector is not provider_copy
    assert isinstance(provider_copy, providers.Selector)

    with switch.override("stdin"):
        assert selector() is sys.stdin

    with switch.override("stdout"):
        assert selector() is sys.stdout

    with switch.override("stderr"):
        assert selector() is sys.stderr


def test_repr(selector, switch):
    assert "<dependency_injector.providers.Selector({0}".format(repr(switch)) in repr(selector)
    assert "one={0}".format(repr(selector.one)) in repr(selector)
    assert "two={0}".format(repr(selector.two)) in repr(selector)
    assert "at {0}".format(hex(id(selector))) in repr(selector)



"""Provider's traversal tests."""

from dependency_injector import providers


def test_traverse_cycled_graph():
    provider1 = providers.Provider()

    provider2 = providers.Provider()
    provider2.override(provider1)

    provider3 = providers.Provider()
    provider3.override(provider2)

    provider1.override(provider3)  # Cycle: provider3 -> provider2 -> provider1 -> provider3

    all_providers = list(providers.traverse(provider1))

    assert len(all_providers) == 3
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provider3 in all_providers


def test_traverse_types_filtering():
    provider1 = providers.Resource(dict)
    provider2 = providers.Resource(dict)
    provider3 = providers.Provider()

    provider = providers.Provider()

    provider.override(provider1)
    provider.override(provider2)
    provider.override(provider3)

    all_providers = list(providers.traverse(provider, types=[providers.Resource]))

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers



"""Factory provider async mode tests."""

import asyncio

from dependency_injector import containers, providers
from pytest import mark, raises

from .common import RESOURCE1, RESOURCE2, Client, Service, BaseContainer, Container, init_resource


@mark.asyncio
async def test_args_injection():
    class ContainerWithArgs(BaseContainer):
        client = providers.Factory(
            Client,
            BaseContainer.resource1,
            BaseContainer.resource2,
        )

        service = providers.Factory(
            Service,
            client,
        )

    container = ContainerWithArgs()

    client1 = await container.client()
    client2 = await container.client()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service()
    service2 = await container.service()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1.client is not service2.client


@mark.asyncio
async def test_kwargs_injection():
    class ContainerWithKwArgs(Container):
        ...

    container = ContainerWithKwArgs()

    client1 = await container.client()
    client2 = await container.client()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service()
    service2 = await container.service()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1.client is not service2.client


@mark.asyncio
async def test_context_kwargs_injection():
    resource2_extra = object()

    container = Container()

    client1 = await container.client(resource2=resource2_extra)
    client2 = await container.client(resource2=resource2_extra)

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is resource2_extra

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is resource2_extra


@mark.asyncio
async def test_args_kwargs_injection():
    class ContainerWithArgsAndKwArgs(BaseContainer):
        client = providers.Factory(
            Client,
            BaseContainer.resource1,
            resource2=BaseContainer.resource2,
        )

        service = providers.Factory(
            Service,
            client=client,
        )

    container = ContainerWithArgsAndKwArgs()

    client1 = await container.client()
    client2 = await container.client()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service()
    service2 = await container.service()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1.client is not service2.client


@mark.asyncio
async def test_async_provider_with_async_injections():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/368
    async def async_client_provider():
        return {"client": "OK"}

    async def async_service(client):
        return {"service": "OK", "client": client}

    class Container(containers.DeclarativeContainer):
        client = providers.Factory(async_client_provider)
        service = providers.Factory(async_service, client=client)

    container = Container()
    service = await container.service()

    assert service == {"service": "OK", "client": {"client": "OK"}}


@mark.asyncio
async def test_with_awaitable_injection():
    class SomeResource:
        def __await__(self):
            raise RuntimeError("Should never happen")

    async def init_resource():
        yield SomeResource()

    class Service:
        def __init__(self, resource) -> None:
            self.resource = resource

    class Container(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource)
        service = providers.Factory(Service, resource=resource)

    container = Container()

    assert isinstance(container.service(), asyncio.Future)
    assert isinstance(container.resource(), asyncio.Future)

    resource = await container.resource()
    service = await container.service()

    assert isinstance(resource, SomeResource)
    assert isinstance(service.resource, SomeResource)
    assert service.resource is resource


@mark.asyncio
async def test_with_awaitable_injection_and_with_init_resources_call():
    class SomeResource:
        def __await__(self):
            raise RuntimeError("Should never happen")

    async def init_resource():
        yield SomeResource()

    class Service:
        def __init__(self, resource) -> None:
            self.resource = resource

    class Container(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource)
        service = providers.Factory(Service, resource=resource)

    container = Container()

    await container.init_resources()
    assert isinstance(container.service(), asyncio.Future)
    assert isinstance(container.resource(), asyncio.Future)

    resource = await container.resource()
    service = await container.service()

    assert isinstance(resource, SomeResource)
    assert isinstance(service.resource, SomeResource)
    assert service.resource is resource


@mark.asyncio
async def test_injection_error():
    async def init_resource():
        raise Exception("Something went wrong")

    class Container(containers.DeclarativeContainer):
        resource_with_error = providers.Resource(init_resource)

        client = providers.Factory(
            Client,
            resource1=resource_with_error,
            resource2=None,
        )

    container = Container()

    with raises(Exception, match="Something went wrong"):
        await container.client()


@mark.asyncio
async def test_injection_runtime_error_async_provides():
    async def create_client(*args,  **kwargs):
        raise Exception("Something went wrong")

    class Container(BaseContainer):
        client = providers.Factory(
            create_client,
            resource1=BaseContainer.resource1,
            resource2=None,
        )

    container = Container()

    with raises(Exception, match="Something went wrong"):
        await container.client()


@mark.asyncio
async def test_injection_call_error_async_provides():
    async def create_client():  # <-- no args defined
        ...

    class Container(BaseContainer):
        client = providers.Factory(
            create_client,
            resource1=BaseContainer.resource1,
            resource2=None,
        )

    container = Container()

    with raises(TypeError) as exception_info:
        await container.client()
    assert "create_client() got" in str(exception_info.value)
    assert "unexpected keyword argument" in str(exception_info.value)


@mark.asyncio
async def test_attributes_injection():
    class ContainerWithAttributes(BaseContainer):
        client = providers.Factory(
            Client,
            BaseContainer.resource1,
            resource2=None,
        )
        client.add_attributes(resource2=BaseContainer.resource2)

        service = providers.Factory(
            Service,
            client=None,
        )
        service.add_attributes(client=client)

    container = ContainerWithAttributes()

    client1 = await container.client()
    client2 = await container.client()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service()
    service2 = await container.service()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1.client is not service2.client


@mark.asyncio
async def test_attributes_injection_attribute_error():
    class ClientWithException(Client):
        @property
        def attribute_set_error(self):
            return None

        @attribute_set_error.setter
        def attribute_set_error(self, value):
            raise Exception("Something went wrong")

    class Container(BaseContainer):
        client = providers.Factory(
            ClientWithException,
            resource1=BaseContainer.resource1,
            resource2=BaseContainer.resource2,
        )
        client.add_attributes(attribute_set_error=123)

    container = Container()

    with raises(Exception, match="Something went wrong"):
        await container.client()


@mark.asyncio
async def test_attributes_injection_runtime_error():
    async def init_resource():
        raise Exception("Something went wrong")

    class Container(containers.DeclarativeContainer):
        resource = providers.Resource(init_resource)

        client = providers.Factory(
            Client,
            resource1=None,
            resource2=None,
        )
        client.add_attributes(resource1=resource)
        client.add_attributes(resource2=resource)

    container = Container()

    with raises(Exception, match="Something went wrong"):
        await container.client()


@mark.asyncio
async def test_async_instance_and_sync_attributes_injection():
    class ContainerWithAttributes(BaseContainer):
        resource1 = providers.Resource(init_resource, providers.Object(RESOURCE1))

        client = providers.Factory(
            Client,
            BaseContainer.resource1,
            resource2=None,
        )
        client.add_attributes(resource2=providers.Object(RESOURCE2))

        service = providers.Factory(
            Service,
            client=None,
        )
        service.add_attributes(client=client)

    container = ContainerWithAttributes()

    client1 = await container.client()
    client2 = await container.client()

    assert isinstance(client1, Client)
    assert client1.resource1 is RESOURCE1
    assert client1.resource2 is RESOURCE2

    assert isinstance(client2, Client)
    assert client2.resource1 is RESOURCE1
    assert client2.resource2 is RESOURCE2

    service1 = await container.service()
    service2 = await container.service()

    assert isinstance(service1, Service)
    assert isinstance(service1.client, Client)
    assert service1.client.resource1 is RESOURCE1
    assert service1.client.resource2 is RESOURCE2

    assert isinstance(service2, Service)
    assert isinstance(service2.client, Client)
    assert service2.client.resource1 is RESOURCE1
    assert service2.client.resource2 is RESOURCE2

    assert service1.client is not service2.client



"""Auto loader tests."""

import contextlib
import importlib

from dependency_injector.wiring import register_loader_containers, unregister_loader_containers
from pytest import fixture

from samples.wiringstringids import module
from samples.wiringstringids.service import Service
from samples.wiringstringids.container import Container


@fixture
def container():
    container = Container()

    yield container

    with contextlib.suppress(ValueError):
        unregister_loader_containers(container)
    container.unwire()
    importlib.reload(module)


def test_register_container(container: Container) -> None:
    register_loader_containers(container)
    importlib.reload(module)

    service = module.test_function()
    assert isinstance(service, Service)



from typing import Union, Optional, Callable as _Callable, Any

from flask import request as flask_request
from dependency_injector import providers


request: providers.Object[flask_request]


class Application(providers.Singleton): ...


class Extension(providers.Singleton): ...


class View(providers.Callable):
    def as_view(self) -> _Callable[..., Any]: ...


class ClassBasedView(providers.Factory):
    def as_view(self, name: str) -> _Callable[..., Any]: ...


def as_view(provider: Union[View, ClassBasedView], name: Optional[str] = None) -> _Callable[..., Any]: ...



"""Container provider tests."""

import copy

from dependency_injector import containers, providers, errors
from pytest import raises


TEST_VALUE_1 = "core_section_value1"
TEST_CONFIG_1 = {
    "core": {
        "section": {
            "value": TEST_VALUE_1,
        },
    },
}

TEST_VALUE_2 = "core_section_value2"
TEST_CONFIG_2 = {
    "core": {
        "section": {
            "value": TEST_VALUE_2,
        },
    },
}


def _copied(value):
    return copy.deepcopy(value)


class Core(containers.DeclarativeContainer):
    config = providers.Configuration("core")
    value_getter = providers.Callable(lambda _: _, config.section.value)


class Application(containers.DeclarativeContainer):
    config = providers.Configuration("config")
    core = providers.Container(Core, config=config.core)
    dict_factory = providers.Factory(dict, value=core.value_getter)


def test():
    application = Application(config=_copied(TEST_CONFIG_1))
    assert application.dict_factory() == {"value": TEST_VALUE_1}


def test_double_override():
    application = Application()
    application.config.override(_copied(TEST_CONFIG_1))
    application.config.override(_copied(TEST_CONFIG_2))
    assert application.dict_factory() == {"value": TEST_VALUE_2}


def test_override():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/354
    class D(containers.DeclarativeContainer):
        foo = providers.Object("foo")

    class A(containers.DeclarativeContainer):
        d = providers.DependenciesContainer()
        bar = providers.Callable(lambda f: f + "++", d.foo.provided)

    class B(containers.DeclarativeContainer):
        d = providers.Container(D)

        a = providers.Container(A, d=d)

    b = B(d=D())
    result = b.a().bar()
    assert result == "foo++"


def test_override_not_root_provider():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/379
    class NestedContainer(containers.DeclarativeContainer):
        settings = providers.Configuration()

        print_settings = providers.Callable(
            lambda s: s,
            settings,
        )

    class TestContainer(containers.DeclarativeContainer):
        settings = providers.Configuration()

        root_container = providers.Container(
            NestedContainer,
            settings=settings,
        )

        not_root_container = providers.Selector(
            settings.container,
            using_factory=providers.Factory(
                NestedContainer,
                settings=settings,
            ),
            using_container=providers.Container(
                NestedContainer,
                settings=settings,
            )
        )

    container_using_factory = TestContainer(settings=dict(
        container="using_factory",
        foo="bar"
    ))
    assert container_using_factory.root_container().print_settings() == {"container": "using_factory", "foo": "bar"}
    assert container_using_factory.not_root_container().print_settings() == {"container": "using_factory", "foo": "bar"}

    container_using_container = TestContainer(settings=dict(
        container="using_container",
        foo="bar"
    ))
    assert container_using_container.root_container().print_settings() == {"container": "using_container", "foo": "bar"}
    assert container_using_container.not_root_container().print_settings() == {"container": "using_container", "foo": "bar"}


def test_override_by_not_a_container():
    provider = providers.Container(Core)

    with raises(errors.Error):
        provider.override(providers.Object("foo"))


def test_lazy_overriding():
    # See: https://github.com/ets-labs/python-dependency-injector/issues/354
    class D(containers.DeclarativeContainer):
        foo = providers.Object("foo")

    class A(containers.DeclarativeContainer):
        d = providers.DependenciesContainer()
        bar = providers.Callable(lambda f: f + "++", d.foo.provided)

    class B(containers.DeclarativeContainer):
        d = providers.DependenciesContainer()

        a = providers.Container(A, d=d)

    b = B(d=D())
    result = b.a().bar()
    assert result == "foo++"


def test_lazy_overriding_deep():
    # Extended version of test_lazy_overriding()
    class D(containers.DeclarativeContainer):
        foo = providers.Object("foo")

    class C(containers.DeclarativeContainer):
        d = providers.DependenciesContainer()
        bar = providers.Callable(lambda f: f + "++", d.foo.provided)

    class A(containers.DeclarativeContainer):
        d = providers.DependenciesContainer()
        c = providers.Container(C, d=d)

    class B(containers.DeclarativeContainer):
        d = providers.DependenciesContainer()

        a = providers.Container(A, d=d)

    b = B(d=D())
    result = b.a().c().bar()
    assert result == "foo++"


def test_reset_last_overriding():
    application = Application(config=_copied(TEST_CONFIG_1))
    application.core.override(Core(config=_copied(TEST_CONFIG_2["core"])))

    application.core.reset_last_overriding()

    assert application.dict_factory() == {"value": TEST_VALUE_1}


def test_reset_last_overriding_only_overridden():
    application = Application(config=_copied(TEST_CONFIG_1))
    application.core.override(providers.DependenciesContainer(config=_copied(TEST_CONFIG_2["core"])))

    application.core.reset_last_overriding()

    assert application.dict_factory() == {"value": TEST_VALUE_1}


def test_override_context_manager():
    application = Application(config=_copied(TEST_CONFIG_1))
    overriding_core = Core(config=_copied(TEST_CONFIG_2["core"]))

    with application.core.override(overriding_core) as context_core:
        assert application.dict_factory() == {"value": TEST_VALUE_2}
        assert context_core() is overriding_core

    assert application.dict_factory() == {"value": TEST_VALUE_1}


def test_reset_override():
    application = Application(config=_copied(TEST_CONFIG_1))
    application.core.override(Core(config=_copied(TEST_CONFIG_2["core"])))

    application.core.reset_override()

    assert application.dict_factory() == {"value": None}


def test_reset_override_only_overridden():
    application = Application(config=_copied(TEST_CONFIG_1))
    application.core.override(providers.DependenciesContainer(config=_copied(TEST_CONFIG_2["core"])))

    application.core.reset_override()

    assert application.dict_factory() == {"value": None}


def test_assign_parent():
    parent = providers.DependenciesContainer()
    provider = providers.Container(Core)

    provider.assign_parent(parent)

    assert provider.parent is parent


def test_parent_name():
    container = containers.DynamicContainer()
    provider = providers.Container(Core)
    container.name = provider
    assert provider.parent_name == "name"


def test_parent_name_with_deep_parenting():
    provider = providers.Container(Core)
    container = providers.DependenciesContainer(name=provider)
    _ = providers.DependenciesContainer(container=container)
    assert provider.parent_name == "container.name"


def test_parent_name_is_none():
    provider = providers.Container(Core)
    assert provider.parent_name is None


def test_parent_deepcopy():
    container = containers.DynamicContainer()
    provider = providers.Container(Core)
    container.name = provider

    copied = providers.deepcopy(container)

    assert container.name.parent is container
    assert copied.name.parent is copied

    assert container is not copied
    assert container.name is not copied.name
    assert container.name.parent is not copied.name.parent


def test_resolve_provider_name():
    container = providers.Container(Core)
    assert container.resolve_provider_name(container.value_getter) == "value_getter"


def test_resolve_provider_name_no_provider():
    container = providers.Container(Core)
    with raises(errors.Error):
        container.resolve_provider_name(providers.Provider())



from typing import Optional

from dependency_injector import providers

# Test 1: to check the return type
provider1 = providers.Delegate(providers.Provider())
var1: providers.Provider = provider1()

# Test 2: to check the return type with await
provider2 = providers.Delegate(providers.Provider())
async def _async2() -> None:
    var1: providers.Provider = await provider2()  # type: ignore
    var2: providers.Provider = await provider2.async_()

# Test 3: to check class type from provider
provider3 = providers.Delegate(providers.Provider())
provided_provides: Optional[providers.Provider] = provider3.provides



"""Dependency Injector Factory providers benchmark."""

import time

from dependency_injector import providers


N = 1000000


class A(object):
    pass


class B(object):
    pass


class C(object):
    pass


class Test(object):
    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c


# Testing Factory provider

test_factory_provider = providers.Factory(
    Test,
    a=providers.Factory(A),
    b=providers.Factory(B),
    c=providers.Factory(C),
)

start = time.time()
for _ in range(1, N):
    test_factory_provider()
finish = time.time()

print(finish - start)


# Testing simple analog

def test_simple_factory_provider():
    return Test(a=A(), b=B(), c=C())


start = time.time()
for _ in range(1, N):
    test_simple_factory_provider()
finish = time.time()

print(finish - start)

# ------
# Result
# ------
#
# Python 2.7
#
# $ python tests/performance/factory_benchmark_1.py
# 0.87456202507
# 0.879760980606
#
# $ python tests/performance/factory_benchmark_1.py
# 0.949290990829
# 0.853044986725
#
# $ python tests/performance/factory_benchmark_1.py
# 0.964688062668
# 0.857432842255
#
# Python 3.7.0
#
# $ python tests/performance/factory_benchmark_1.py
# 1.1037120819091797
# 0.999565839767456
#
# $ python tests/performance/factory_benchmark_1.py
# 1.0855588912963867
# 1.0008318424224854
#
# $ python tests/performance/factory_benchmark_1.py
# 1.0706679821014404
# 1.0106139183044434



"""Delegate provider tests."""

from dependency_injector import providers, errors
from pytest import fixture, raises


@fixture
def provider():
    return providers.Provider()


@fixture
def delegate(provider):
    return providers.Delegate(provider)


def test_is_provider(delegate):
    assert providers.is_provider(delegate) is True


def test_init_optional_provides(provider):
    delegate = providers.Delegate()
    delegate.set_provides(provider)
    assert delegate.provides is provider
    assert delegate() is provider


def test_set_provides_returns_self(delegate, provider):
    assert delegate.set_provides(provider) is delegate


def test_init_with_not_provider():
    with raises(errors.Error):
        providers.Delegate(object())


def test_call(delegate, provider):
    delegated1 = delegate()
    delegated2 = delegate()

    assert delegated1 is provider
    assert delegated2 is provider


def test_repr(delegate, provider):
    assert repr(delegate) == (
        "<dependency_injector.providers."
        "Delegate({0}) at {1}>".format(repr(provider), hex(id(delegate)))
    )



"""Configuration.from_ini() with environment variables interpolation tests."""

import os

from pytest import mark, raises


def test_env_variable_interpolation(config, ini_config_file_3):
    config.from_ini(ini_config_file_3)

    assert config() == {
        "section1": {
            "value1": "test-value",
            "value2": "test-path/path",
        },
    }
    assert config.section1() == {
        "value1": "test-value",
        "value2": "test-path/path",
    }
    assert config.section1.value1() == "test-value"
    assert config.section1.value2() == "test-path/path"


def test_missing_envs_not_required(config, ini_config_file_3):
    del os.environ["CONFIG_TEST_ENV"]
    del os.environ["CONFIG_TEST_PATH"]

    config.from_ini(ini_config_file_3)

    assert config() == {
        "section1": {
            "value1": "",
            "value2": "/path",
        },
    }
    assert config.section1() == {
        "value1": "",
        "value2": "/path",
    }
    assert config.section1.value1() == ""
    assert config.section1.value2() == "/path"


def test_missing_envs_required(config, ini_config_file_3):
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "undefined=${UNDEFINED}\n"
        )
    with raises(ValueError, match="Missing required environment variable \"UNDEFINED\""):
        config.from_ini(ini_config_file_3, envs_required=True)


@mark.parametrize("config_type", ["strict"])
def test_missing_envs_strict_mode(config, ini_config_file_3):
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "undefined=${UNDEFINED}\n"
        )
    with raises(ValueError, match="Missing required environment variable \"UNDEFINED\""):
        config.from_ini(ini_config_file_3)


@mark.parametrize("config_type", ["strict"])
def test_missing_envs_not_required_in_strict_mode(config, ini_config_file_3):
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "undefined=${UNDEFINED}\n"
        )
    config.from_ini(ini_config_file_3, envs_required=False)
    assert config.section.undefined() == ""


def test_option_missing_envs_not_required(config, ini_config_file_3):
    del os.environ["CONFIG_TEST_ENV"]
    del os.environ["CONFIG_TEST_PATH"]

    config.option.from_ini(ini_config_file_3)

    assert config.option() == {
        "section1": {
            "value1": "",
            "value2": "/path",
        },
    }
    assert config.option.section1() == {
        "value1": "",
        "value2": "/path",
    }
    assert config.option.section1.value1() == ""
    assert config.option.section1.value2() == "/path"


def test_option_missing_envs_required(config, ini_config_file_3):
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "undefined=${UNDEFINED}\n"
        )
    with raises(ValueError, match="Missing required environment variable \"UNDEFINED\""):
        config.option.from_ini(ini_config_file_3, envs_required=True)


@mark.parametrize("config_type", ["strict"])
def test_option_missing_envs_not_required_in_strict_mode(config, ini_config_file_3):
    config.override({"option": {}})
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "undefined=${UNDEFINED}\n"
        )
    config.option.from_ini(ini_config_file_3, envs_required=False)
    assert config.option.section.undefined() == ""


@mark.parametrize("config_type", ["strict"])
def test_option_missing_envs_strict_mode(config, ini_config_file_3):
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "undefined=${UNDEFINED}\n"
        )
    with raises(ValueError, match="Missing required environment variable \"UNDEFINED\""):
        config.option.from_ini(ini_config_file_3)


def test_default_values(config, ini_config_file_3):
    with open(ini_config_file_3, "w") as file:
        file.write(
            "[section]\n"
            "defined_with_default=${DEFINED:default}\n"
            "undefined_with_default=${UNDEFINED:default}\n"
            "complex=${DEFINED}/path/${DEFINED:default}/${UNDEFINED}/${UNDEFINED:default}\n"
        )

    config.from_ini(ini_config_file_3)

    assert config.section() == {
        "defined_with_default": "defined",
        "undefined_with_default": "default",
        "complex": "defined/path/defined//default",
    }



"""Tests for wiring to module as package."""

from pytest import fixture

from samples.wiring import module
from samples.wiring.service import Service
from samples.wiring.container import Container


@fixture
def container():
    container = Container()
    yield container
    container.unwire()


def test_module_as_package_wiring(container: Container):
    # See: https://github.com/ets-labs/python-dependency-injector/issues/481
    container.wire(packages=[module])
    assert isinstance(module.service, Service)



"""`Dependency injector` setup script."""

import os

from Cython.Build import cythonize
from Cython.Compiler import Options
from setuptools import Extension, setup

debug = os.environ.get("DEPENDENCY_INJECTOR_DEBUG_MODE") == "1"
defined_macros = []
compiler_directives = {
    "language_level": 3,
    "profile": debug,
    "linetrace": debug,
}
Options.annotate = debug

# Adding debug options:
if debug:
    defined_macros.extend(
        [
            ("CYTHON_TRACE", "1"),
            ("CYTHON_TRACE_NOGIL", "1"),
            ("CYTHON_CLINE_IN_TRACEBACK", "1"),
        ]
    )


setup(
    ext_modules=cythonize(
        [
            Extension(
                "*",
                ["src/**/*.pyx"],
                define_macros=defined_macros,
            ),
        ],
        annotate=debug,
        show_all_warnings=True,
        compiler_directives=compiler_directives,
    ),
)



"""Tests for container self provider."""

from dependency_injector import containers, providers, errors
from pytest import raises


def test_self():
    def call_bar(container):
        return container.bar()

    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = providers.Callable(call_bar, __self__)
        bar = providers.Object("hello")

    container = Container()
    assert container.foo() is "hello"


def test_self_attribute_implicit():
    class Container(containers.DeclarativeContainer):
        pass

    container = Container()
    assert container.__self__() is container


def test_self_attribute_explicit():
    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()

    container = Container()
    assert container.__self__() is container


def test_single_self():
    with raises(errors.Error):
        class Container(containers.DeclarativeContainer):
            self1 = providers.Self()
            self2 = providers.Self()


def test_self_attribute_alt_name_implicit():
    class Container(containers.DeclarativeContainer):
        foo = providers.Self()

    container = Container()

    assert container.__self__ is container.foo
    assert set(container.__self__.alt_names) == {"foo"}


def test_self_attribute_alt_name_explicit_1():
    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = __self__
        bar = __self__

    container = Container()

    assert container.__self__ is container.foo
    assert container.__self__ is container.bar
    assert set(container.__self__.alt_names) == {"foo", "bar"}


def test_self_attribute_alt_name_explicit_2():
    class Container(containers.DeclarativeContainer):
        foo = providers.Self()
        bar = foo

    container = Container()

    assert container.__self__ is container.foo
    assert container.__self__ is container.bar
    assert set(container.__self__.alt_names) == {"foo", "bar"}


def test_providers_attribute_1():
    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = __self__
        bar = __self__

    container = Container()

    assert container.providers == {}
    assert Container.providers == {}


def test_providers_attribute_2():
    class Container(containers.DeclarativeContainer):
        foo = providers.Self()
        bar = foo

    container = Container()

    assert container.providers == {}
    assert Container.providers == {}


def test_container_multiple_instances():
    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()

    container1 = Container()
    container2 = Container()

    assert container1 is not container2
    assert container1.__self__() is container1
    assert container2.__self__() is container2


def test_deepcopy():
    def call_bar(container):
        return container.bar()

    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = providers.Callable(call_bar, __self__)
        bar = providers.Object("hello")

    container1 = Container()
    container2 = providers.deepcopy(container1)
    container1.bar.override("bye")

    assert container1.foo() == "bye"
    assert container2.foo() == "hello"


def test_deepcopy_alt_names_1():
    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = __self__
        bar = foo

    container1 = Container()
    container2 = providers.deepcopy(container1)

    assert container2.__self__() is container2
    assert container2.foo() is container2
    assert container2.bar() is container2


def test_deepcopy_alt_names_2():
    class Container(containers.DeclarativeContainer):
        self = providers.Self()

    container1 = Container()
    container2 = providers.deepcopy(container1)

    assert container2.__self__() is container2
    assert container2.self() is container2


def test_deepcopy_no_self_dependencies():
    class Container(containers.DeclarativeContainer):
        __self__ = providers.Self()

    container1 = Container()
    container2 = providers.deepcopy(container1)

    assert container1 is not container2
    assert container1.__self__ is not container2.__self__
    assert container1.__self__() is container1
    assert container2.__self__() is container2


def test_with_container_provider():
    def call_bar(container):
        return container.bar()

    class SubContainer(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = providers.Callable(call_bar, __self__)
        bar = providers.Object("hello")

    class Container(containers.DeclarativeContainer):
        sub_container = providers.Container(SubContainer)

        baz = providers.Callable(lambda value: value, sub_container.foo)

    container = Container()
    assert container.baz() == "hello"


def test_with_container_provider_overriding():
    def call_bar(container):
        return container.bar()

    class SubContainer(containers.DeclarativeContainer):
        __self__ = providers.Self()
        foo = providers.Callable(call_bar, __self__)
        bar = providers.Object("hello")

    class Container(containers.DeclarativeContainer):
        sub_container = providers.Container(SubContainer, bar="bye")

        baz = providers.Callable(lambda value: value, sub_container.foo)

    container = Container()
    assert container.baz() == "bye"


def test_with_container_provider_self():
    class SubContainer(containers.DeclarativeContainer):
        __self__ = providers.Self()

    class Container(containers.DeclarativeContainer):
        sub_container = providers.Container(SubContainer)

    container = Container()

    assert container.__self__() is container
    assert container.sub_container().__self__() is container.sub_container()




"""Provider tests."""

import warnings

from dependency_injector import providers, errors
from pytest import fixture, raises


@fixture
def provider():
    return providers.Provider()


def test_is_provider(provider):
    assert providers.is_provider(provider) is True


def test_call(provider):
    with raises(NotImplementedError):
        provider()


def test_delegate(provider):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        delegate1 = provider.delegate()
        delegate2 = provider.delegate()

    assert isinstance(delegate1, providers.Delegate)
    assert delegate1() is provider

    assert isinstance(delegate2, providers.Delegate)
    assert delegate2() is provider

    assert delegate1 is not delegate2


def test_provider(provider):
    delegate1 = provider.provider

    assert isinstance(delegate1, providers.Delegate)
    assert delegate1() is provider

    delegate2 = provider.provider

    assert isinstance(delegate2, providers.Delegate)
    assert delegate2() is provider

    assert delegate1 is not delegate2


def test_override(provider):
    overriding_provider = providers.Provider()
    provider.override(overriding_provider)
    assert provider.overridden == (overriding_provider,)
    assert provider.last_overriding is overriding_provider


def test_double_override(provider):
    overriding_provider1 = providers.Object(1)
    overriding_provider2 = providers.Object(2)

    provider.override(overriding_provider1)
    overriding_provider1.override(overriding_provider2)

    assert provider() == overriding_provider2()


def test_overriding_context(provider):
    overriding_provider = providers.Provider()
    with provider.override(overriding_provider):
        assert provider.overridden == (overriding_provider,)
    assert provider.overridden == tuple()
    assert not provider.overridden


def test_override_with_itself(provider):
    with raises(errors.Error):
        provider.override(provider)


def test_override_with_not_provider(provider):
    obj = object()
    provider.override(obj)
    assert provider() is obj


def test_reset_last_overriding(provider):
    overriding_provider1 = providers.Provider()
    overriding_provider2 = providers.Provider()

    provider.override(overriding_provider1)
    provider.override(overriding_provider2)

    assert provider.overridden[-1] is overriding_provider2
    assert provider.last_overriding is overriding_provider2

    provider.reset_last_overriding()
    assert provider.overridden[-1] is overriding_provider1
    assert provider.last_overriding is overriding_provider1

    provider.reset_last_overriding()
    assert provider.overridden == tuple()
    assert not provider.overridden
    assert provider.last_overriding is None


def test_reset_last_overriding_of_not_overridden_provider(provider):
    with raises(errors.Error):
        provider.reset_last_overriding()


def test_reset_override(provider):
    overriding_provider = providers.Provider()
    provider.override(overriding_provider)

    assert provider.overridden
    assert provider.overridden == (overriding_provider,)

    provider.reset_override()

    assert provider.overridden == tuple()


def test_deepcopy(provider):
    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider, providers.Provider)


def test_deepcopy_from_memo(provider):
    provider_copy_memo = providers.Provider()
    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})
    assert provider_copy is provider_copy_memo


def test_deepcopy_overridden(provider):
    overriding_provider = providers.Provider()

    provider.override(overriding_provider)

    provider_copy = providers.deepcopy(provider)
    overriding_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert isinstance(provider, providers.Provider)

    assert overriding_provider is not overriding_provider_copy
    assert isinstance(overriding_provider_copy, providers.Provider)


def test_repr(provider):
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Provider() at {0}>".format(hex(id(provider)))
    )



"""Configuration(json_files=[...]) tests."""

import json

from dependency_injector import providers
from pytest import fixture, mark, raises


@fixture
def config(config_type, json_config_file_1, json_config_file_2):
    if config_type == "strict":
        return providers.Configuration(strict=True)
    elif config_type == "default":
        return providers.Configuration(json_files=[json_config_file_1, json_config_file_2])
    else:
        raise ValueError("Undefined config type \"{0}\"".format(config_type))


def test_load(config):
    config.load()

    assert config() == {
        "section1": {
            "value1": 11,
            "value11": 11,
        },
        "section2": {
            "value2": 2,
        },
        "section3": {
            "value3": 3,
        },
    }
    assert config.section1() == {"value1": 11, "value11": 11}
    assert config.section1.value1() == 11
    assert config.section1.value11() == 11
    assert config.section2() == {"value2": 2}
    assert config.section2.value2() == 2
    assert config.section3() == {"value3": 3}
    assert config.section3.value3() == 3


def test_get_files(config, json_config_file_1, json_config_file_2):
    assert config.get_json_files() == [json_config_file_1, json_config_file_2]


def test_set_files(config):
    config.set_json_files(["file1.json", "file2.json"])
    assert config.get_json_files() == ["file1.json", "file2.json"]


def test_copy(config, json_config_file_1, json_config_file_2):
    config_copy = providers.deepcopy(config)
    assert config_copy.get_json_files() == [json_config_file_1, json_config_file_2]


def test_file_does_not_exist(config):
    config.set_json_files(["./does_not_exist.json"])
    config.load()
    assert config() == {}


@mark.parametrize("config_type", ["strict"])
def test_file_does_not_exist_strict_mode(config):
    config.set_json_files(["./does_not_exist.json"])
    with raises(IOError):
        config.load()
    assert config() == {}


def test_required_file_does_not_exist(config):
    config.set_json_files(["./does_not_exist.json"])
    with raises(IOError):
        config.load(required=True)


@mark.parametrize("config_type", ["strict"])
def test_not_required_file_does_not_exist_strict_mode(config):
    config.set_json_files(["./does_not_exist.json"])
    config.load(required=False)
    assert config() == {}


def test_missing_envs_required(config, json_config_file_3):
    with open(json_config_file_3, "w") as file:
        file.write(
            json.dumps(
                {
                    "section": {
                        "undefined": "${UNDEFINED}",
                    },
                },
            ),
        )
    config.set_json_files([json_config_file_3])
    with raises(ValueError, match="Missing required environment variable \"UNDEFINED\""):
        config.load(envs_required=True)


@mark.parametrize("config_type", ["strict"])
def test_missing_envs_not_required_in_strict_mode(config, json_config_file_3):
    with open(json_config_file_3, "w") as file:
        file.write(
            json.dumps(
                {
                    "section": {
                        "undefined": "${UNDEFINED}",
                    },
                },
            ),
        )
    config.set_json_files([json_config_file_3])
    config.load(envs_required=False)
    assert config.section.undefined() == ""



"""AbstractCallable provider tests."""

from dependency_injector import providers, errors
from pytest import raises

from .common import example


def test_inheritance():
    assert isinstance(providers.AbstractCallable(example), providers.Callable)


def test_call_overridden_by_callable():
    def _abstract_example():
        pass

    provider = providers.AbstractCallable(_abstract_example)
    provider.override(providers.Callable(example))

    assert provider(1, 2, 3, 4) == (1, 2, 3, 4)


def test_call_overridden_by_delegated_callable():
    def _abstract_example():
        pass

    provider = providers.AbstractCallable(_abstract_example)
    provider.override(providers.DelegatedCallable(example))

    assert provider(1, 2, 3, 4) == (1, 2, 3, 4)


def test_call_not_overridden():
    provider = providers.AbstractCallable(example)
    with raises(errors.Error):
        provider(1, 2, 3, 4)


def test_override_by_not_callable():
    provider = providers.AbstractCallable(example)
    with raises(errors.Error):
        provider.override(providers.Factory(object))


def test_provide_not_implemented():
    provider = providers.AbstractCallable(example)
    with raises(NotImplementedError):
        provider._provide((1, 2, 3, 4), dict())


def test_repr():
    provider = providers.AbstractCallable(example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "AbstractCallable({0}) at {1}>".format(repr(example), hex(id(provider)))
    )



"""SingletonDelegate provider tests."""

from dependency_injector import providers, errors
from pytest import fixture, raises


@fixture
def provider():
    return providers.Singleton(object)


@fixture
def delegate(provider):
    return providers.SingletonDelegate(provider)


def test_is_delegate(delegate):
    assert isinstance(delegate, providers.Delegate)


def test_init_with_not_factory():
    with raises(errors.Error):
        providers.SingletonDelegate(providers.Object(object()))



from typing import Callable, Optional, Tuple, Any, Dict, Type

from dependency_injector import providers


class Animal:
    ...


class Cat(Animal):

    @classmethod
    def create(cls) -> Animal:
        return cls()


# Test 1: to check the return type (class)
provider1 = providers.Callable(Cat)
animal1: Animal = provider1(1, 2, 3, b="1", c=2, e=0.0)

# Test 2: to check the return type (class factory method)
provider2 = providers.Callable(Cat.create)
animal2: Animal = provider2()

# Test 3: to check the .override() method
provider3 = providers.Callable(Animal)
with provider3.override(providers.Callable(Cat)):
    provider3()

# Test 4: to check the .args & .kwargs attributes
provider4 = providers.Callable(Animal)
args4: Tuple[Any] = provider4.args
kwargs4: Dict[str, Any] = provider4.kwargs

# Test 5: to check the provided instance interface
provider5 = providers.Callable(Animal)
provided5: providers.ProvidedInstance = provider5.provided
attr_getter5: providers.AttributeGetter = provider5.provided.attr
item_getter5: providers.ItemGetter = provider5.provided["item"]
method_caller: providers.MethodCaller = provider5.provided.method.call(123, arg=324)

# Test 6: to check the DelegatedCallable
provider6 = providers.DelegatedCallable(Cat)
animal6: Animal = provider6(1, 2, 3, b="1", c=2, e=0.0)

# Test 7: to check the AbstractCallable
provider7 = providers.AbstractCallable(Animal)
provider7.override(providers.Callable(Cat))
animal7: Animal = provider7(1, 2, 3, b="1", c=2, e=0.0)

# Test 8: to check the CallableDelegate __init__
provider8 = providers.CallableDelegate(providers.Callable(lambda: None))

# Test 9: to check the return type with await
provider9 = providers.Callable(Cat)
async def _async9() -> None:
    animal1: Animal = await provider9(1, 2, 3, b="1", c=2, e=0.0)  # type: ignore
    animal2: Animal = await provider9.async_(1, 2, 3, b="1", c=2, e=0.0)

# Test 10: to check the .provides
provider10 = providers.Callable(Cat)
provides10: Optional[Callable[..., Cat]] = provider10.provides
assert provides10 is Cat

# Test 11: to check the .provides for explicit typevar
provider11 = providers.Callable[Animal](Cat)
provides11: Optional[Callable[..., Animal]] = provider11.provides
assert provides11 is Cat

# Test 12: to check string imports
provider12: providers.Callable[dict] = providers.Callable("builtins.dict")
provider12.set_provides("builtins.dict")



"""ThreadSafeSingleton provider async mode tests."""

import asyncio

from dependency_injector import providers
from pytest import mark


@mark.asyncio
async def test_async_mode():
    instance = object()

    async def create_instance():
        return instance

    provider = providers.ThreadSafeSingleton(create_instance)

    instance1 = await provider()
    instance2 = await provider()

    assert instance1 is instance2
    assert instance1 is instance
    assert instance2 is instance


@mark.asyncio
async def test_concurrent_init():
    async def create_instance():
        return object()

    provider = providers.ThreadSafeSingleton(create_instance)

    future_instance1 = provider()
    future_instance2 = provider()

    instance1, instance2 = await asyncio.gather(future_instance1, future_instance2)

    assert instance1 is instance2



"""Callable provider tests."""

import decimal
import sys

from dependency_injector import providers, errors
from pytest import raises, mark

from .common import example


def test_is_provider():
    assert providers.is_provider(providers.Callable(example)) is True


def test_init_with_not_callable():
    with raises(errors.Error):
        providers.Callable(123)


def test_init_optional_provides():
    provider = providers.Callable()
    provider.set_provides(object)
    assert provider.provides is object
    assert isinstance(provider(), object)


def test_set_provides_returns_():
    provider = providers.Callable()
    assert provider.set_provides(object) is provider


@mark.parametrize(
    "str_name,cls",
    [
        ("dependency_injector.providers.Factory", providers.Factory),
        ("decimal.Decimal", decimal.Decimal),
        ("list", list),
        (".common.example", example),
        ("test_is_provider", test_is_provider),
    ],
)
def test_set_provides_string_imports(str_name, cls):
    assert providers.Callable(str_name).provides is cls


def test_provided_instance_provider():
    provider = providers.Callable(example)
    assert isinstance(provider.provided, providers.ProvidedInstance)


def test_call():
    provider = providers.Callable(lambda: True)
    assert provider() is True


def test_call_with_positional_args():
    provider = providers.Callable(example, 1, 2, 3, 4)
    assert provider() == (1, 2, 3, 4)


def test_call_with_keyword_args():
    provider = providers.Callable(example, arg1=1, arg2=2, arg3=3, arg4=4)
    assert provider() == (1, 2, 3, 4)


def test_call_with_positional_and_keyword_args():
    provider = providers.Callable(example, 1, 2, arg3=3, arg4=4)
    assert provider() == (1, 2, 3, 4)


def test_call_with_context_args():
    provider = providers.Callable(example, 1, 2)
    assert provider(3, 4) == (1, 2, 3, 4)


def test_call_with_context_kwargs():
    provider = providers.Callable(example, arg1=1)
    assert provider(arg2=2, arg3=3, arg4=4) == (1, 2, 3, 4)


def test_call_with_context_args_and_kwargs():
    provider = providers.Callable(example, 1)
    assert provider(2, arg3=3, arg4=4) == (1, 2, 3, 4)


def test_fluent_interface():
    provider = providers.Singleton(example) \
        .add_args(1, 2) \
        .add_kwargs(arg3=3, arg4=4)
    assert provider() == (1, 2, 3, 4)


def test_set_args():
    provider = providers.Callable(example) \
        .add_args(1, 2) \
        .set_args(3, 4)
    assert provider.args == (3, 4)


def test_set_kwargs():
    provider = providers.Callable(example) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .set_kwargs(init_arg3=4, init_arg4=5)
    assert provider.kwargs == dict(init_arg3=4, init_arg4=5)


def test_clear_args():
    provider = providers.Callable(example) \
        .add_args(1, 2) \
        .clear_args()
    assert provider.args == tuple()


def test_clear_kwargs():
    provider = providers.Callable(example) \
        .add_kwargs(init_arg3=3, init_arg4=4) \
        .clear_kwargs()
    assert provider.kwargs == dict()


def test_call_overridden():
    provider = providers.Callable(example)

    provider.override(providers.Object((4, 3, 2, 1)))
    provider.override(providers.Object((1, 2, 3, 4)))

    assert provider() == (1, 2, 3, 4)


def test_deepcopy():
    provider = providers.Callable(example)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert provider.provides is provider_copy.provides
    assert isinstance(provider, providers.Callable)


def test_deepcopy_from_memo():
    provider = providers.Callable(example)
    provider_copy_memo = providers.Callable(example)

    provider_copy = providers.deepcopy(provider, memo={id(provider): provider_copy_memo})

    assert provider_copy is provider_copy_memo


def test_deepcopy_args():
    provider = providers.Callable(example)
    dependent_provider1 = providers.Callable(list)
    dependent_provider2 = providers.Callable(dict)

    provider.add_args(dependent_provider1, dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.args[0]
    dependent_provider_copy2 = provider_copy.args[1]

    assert provider.args != provider_copy.args

    assert dependent_provider1.provides is dependent_provider_copy1.provides
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.provides is dependent_provider_copy2.provides
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_kwargs():
    provider = providers.Callable(example)
    dependent_provider1 = providers.Callable(list)
    dependent_provider2 = providers.Callable(dict)

    provider.add_kwargs(a1=dependent_provider1, a2=dependent_provider2)

    provider_copy = providers.deepcopy(provider)
    dependent_provider_copy1 = provider_copy.kwargs["a1"]
    dependent_provider_copy2 = provider_copy.kwargs["a2"]

    assert provider.kwargs != provider_copy.kwargs

    assert dependent_provider1.provides is dependent_provider_copy1.provides
    assert dependent_provider1 is not dependent_provider_copy1

    assert dependent_provider2.provides is dependent_provider_copy2.provides
    assert dependent_provider2 is not dependent_provider_copy2


def test_deepcopy_overridden():
    provider = providers.Callable(example)
    object_provider = providers.Object(object())

    provider.override(object_provider)

    provider_copy = providers.deepcopy(provider)
    object_provider_copy = provider_copy.overridden[0]

    assert provider is not provider_copy
    assert provider.provides is provider_copy.provides
    assert isinstance(provider, providers.Callable)

    assert object_provider is not object_provider_copy
    assert isinstance(object_provider_copy, providers.Object)


def test_deepcopy_with_sys_streams():
    provider = providers.Callable(example)
    provider.add_args(sys.stdin)
    provider.add_kwargs(a2=sys.stdout)

    provider_copy = providers.deepcopy(provider)

    assert provider is not provider_copy
    assert isinstance(provider_copy, providers.Callable)
    assert provider.args[0] is sys.stdin
    assert provider.kwargs["a2"] is sys.stdout


def test_repr():
    provider = providers.Callable(example)
    assert repr(provider) == (
        "<dependency_injector.providers."
        "Callable({0}) at {1}>".format(repr(example), hex(id(provider)))
    )



from dependency_injector import providers


class Container:
    ...


# Test 1: to check the return type
provider1 = providers.Container(Container)
var1: Container = provider1()

# Test 2: to check the getattr
provider2 = providers.Container(Container)
attr: providers.Provider = provider2.attr



"""DependenciesContainer provider traversal tests."""

from dependency_injector import providers


def test_traversal():
    provider = providers.DependenciesContainer()
    all_providers = list(provider.traverse())
    assert len(all_providers) == 0


def test_traversal_default():
    another_provider = providers.Provider()
    provider = providers.DependenciesContainer(default=another_provider)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 1
    assert another_provider in all_providers


def test_traversal_fluent_interface():
    provider = providers.DependenciesContainer()
    provider1 = provider.provider1
    provider2 = provider.provider2

    all_providers = list(provider.traverse())

    assert len(all_providers) == 2
    assert provider1 in all_providers
    assert provider2 in all_providers


def test_traversal_overriding():
    provider1 = providers.Provider()
    provider2 = providers.Provider()
    provider3 = providers.DependenciesContainer(
        provider1=provider1,
        provider2=provider2,
    )

    provider = providers.DependenciesContainer()
    provider.override(provider3)

    all_providers = list(provider.traverse())

    assert len(all_providers) == 5
    assert provider1 in all_providers
    assert provider2 in all_providers
    assert provider3 in all_providers
    assert provider.provider1 in all_providers
    assert provider.provider2 in all_providers
